import torch


def _h_poly(t):
    """Helper function to compute the 'h' polynomial matrix used in the
    cubic spline.

    Args:
        t (Tensor): A 1D tensor representing the normalized x values.

    Returns:
        Tensor: A 2D tensor of size (4, len(t)) representing the 'h' polynomial matrix.

    """

    tt = t[None, :] ** (torch.arange(4, device=t.device)[:, None])
    A = torch.tensor(
        [[1, 0, -3, 2], [0, 1, -2, 1], [0, 0, 3, -2], [0, 0, -1, 1]],
        dtype=t.dtype,
        device=t.device,
    )
    return A @ tt


def cubic_spline_torch(x: torch.Tensor, y: torch.Tensor, xs: torch.Tensor) -> torch.Tensor:
    """Compute the 1D cubic spline interpolation for the given data points
    using PyTorch.

    Args:
        x (Tensor): A 1D tensor representing the x-coordinates of the known data points.
        y (Tensor): A 1D tensor representing the y-coordinates of the known data points.
        xs (Tensor): A 1D tensor representing the x-coordinates of the positions where
                     the cubic spline function should be evaluated.
        extend (str, optional): The method for handling extrapolation, either "const" or "linear".
                                Default is "const".
                                "const": Use the value of the last known data point for extrapolation.
                                "linear": Use linear extrapolation based on the last two known data points.

    Returns:
        Tensor: A 1D tensor representing the interpolated values at the specified positions (xs).

    """
    m = (y[1:] - y[:-1]) / (x[1:] - x[:-1])
    m = torch.cat([m[[0]], (m[1:] + m[:-1]) / 2, m[[-1]]])
    idxs = torch.searchsorted(x[:-1], xs) - 1
    dx = x[idxs + 1] - x[idxs]
    hh = _h_poly((xs - x[idxs]) / dx)
    ret = hh[0] * y[idxs] + hh[1] * m[idxs] * dx + hh[2] * y[idxs + 1] + hh[3] * m[idxs + 1] * dx
    return ret


def spline(R, profR, profI, extend="zeros"):
    """Spline 1d profile function, cubic spline between points up
    to second last point beyond which is linear

    Parameters:
        R: Radii tensor at which to evaluate the sersic function
        profR: radius values for the surface density profile in the same units as R
        profI: surface density values for the surface density profile
    """
    I = cubic_spline_torch(profR, profI, R.view(-1)).reshape(*R.shape)
    if extend == "zeros":
        I[R > profR[-1]] = 0
    elif extend == "const":
        I[R > profR[-1]] = profI[-1]
    else:
        raise ValueError(f"Unknown extend option: {extend}. Use 'zeros' or 'const'.")
    return I



import numpy as np
import torch
from math import prod
from caskade import (
    Module as CModule,
    ActiveStateError,
    ParamConfigurationError,
    FillDynamicParamsArrayError,
)


class Module(CModule):

    def build_params_array_identities(self):
        identities = []
        for param in self.dynamic_params:
            numel = max(1, np.prod(param.shape))
            for i in range(numel):
                identities.append(f"{id(param)}_{i}")
        return identities

    def build_params_array_uncertainty(self):
        uncertainties = []
        for param in self.dynamic_params:
            if param.uncertainty is None:
                uncertainties.append(torch.zeros_like(param.value.flatten()))
            else:
                uncertainties.append(param.uncertainty.flatten())
        return torch.cat(tuple(uncertainties), dim=-1)

    def build_params_array_names(self):
        names = []
        for param in self.dynamic_params:
            numel = max(1, np.prod(param.shape))
            if numel == 1:
                names.append(param.name)
            else:
                for i in range(numel):
                    names.append(f"{param.name}_{i}")
        return names

    def build_params_array_units(self):
        units = []
        for param in self.dynamic_params:
            numel = max(1, np.prod(param.shape))
            for _ in range(numel):
                units.append(param.unit)
        return units

    def fill_dynamic_value_uncertainties(self, uncertainty):
        if self.active:
            raise ActiveStateError(f"Cannot fill dynamic values when Module {self.name} is active")

        dynamic_params = self.dynamic_params

        if uncertainty.shape[-1] == 0:
            return  # No parameters to fill
        # check for batch dimension
        pos = 0
        for param in dynamic_params:
            if not isinstance(param.shape, tuple):
                raise ParamConfigurationError(
                    f"Param {param.name} has no shape. dynamic parameters must have a shape to use Tensor input."
                )
            # Handle scalar parameters
            size = max(1, prod(param.shape))
            try:
                val = uncertainty[..., pos : pos + size].view(param.shape)
                param.uncertainty = val
            except (RuntimeError, IndexError, ValueError, TypeError):
                raise FillDynamicParamsArrayError(self.name, uncertainty, dynamic_params)

            pos += size
        if pos != uncertainty.shape[-1]:
            raise FillDynamicParamsArrayError(self.name, uncertainty, dynamic_params)



from .base import AstroPhotError

__all__ = (
    "InvalidWindow",
    "ConflicingWCS",
    "InvalidData",
    "InvalidImage",
    "InvalidWCS",
)


class InvalidWindow(AstroPhotError):
    """
    Raised whenever a window is misspecified
    """

    ...


class ConflicingWCS(InvalidWindow):
    """
    Raised when windows are compared and have WCS prescriptions which do not agree
    """

    ...


class InvalidData(AstroPhotError):
    """
    Raised when an image object can't determine the data it is holding.
    """

    ...


class InvalidImage(AstroPhotError):
    """
    Raised when an image object cannot be used as given.
    """

    ...


class InvalidWCS(AstroPhotError):
    """
    Raised when the WCS is not appropriate as given.
    """

    ...



import torch
import numpy as np

from ...param import forward


class RadialMixin:
    """This model defines its `brightness(x,y)` function using a radial model.
    Thus the brightness is instead defined as`radial_model(R)`

    More specifically the function is:

    $$x, y = {\\rm transform\\_coordinates}(x, y)$$
    $$R = {\\rm radius\\_metric}(x, y)$$
    $$I(x, y) = {\\rm radial\\_model}(R)$$

    The `transform_coordinates` function depends on the model. In its simplest
    form it simply subtracts the center of the model to re-center the coordinates.

    The `radius_metric` function is also model dependent, in its simplest form
    this is just $R = \\sqrt{x^2 + y^2}$.
    """

    @forward
    def brightness(self, x, y):
        """
        Calculate the brightness at a given point (x, y) based on radial distance from the center.
        """
        x, y = self.transform_coordinates(x, y)
        return self.radial_model(self.radius_metric(x, y))


class WedgeMixin:
    """Defines a model with multiple profiles that form wedges projected from the center.

    model which defines multiple radial models separately along some number of
    wedges projected from the center. These wedges have sharp transitions along boundary angles theta.

    Options:
        symmetric: If True, the model will have symmetry for rotations of pi radians
            and each ray will appear twice on the sky on opposite sides of the model.
            If False, each ray is independent.
        segments: The number of segments to divide the model into. This controls
            how many rays are used in the model. The default is 2
    """

    _model_type = "wedge"
    _options = ("segments", "symmetric")

    def __init__(self, *args, symmetric=True, segments=2, **kwargs):
        super().__init__(*args, **kwargs)
        self.symmetric = symmetric
        self.segments = segments

    def polar_model(self, R, T):
        model = torch.zeros_like(R)
        cycle = np.pi if self.symmetric else 2 * np.pi
        w = cycle / self.segments
        angles = (T + w / 2) % cycle
        v = w * np.arange(self.segments)
        for s in range(self.segments):
            indices = (angles >= v[s]) & (angles < (v[s] + w))
            model[indices] += self.iradial_model(s, R[indices])
        return model

    def brightness(self, x, y):
        x, y = self.transform_coordinates(x, y)
        return self.polar_model(self.radius_metric(x, y), self.angular_metric(x, y))


class RayMixin:
    """Defines a model with multiple profiles along rays projected from the center.

    model which defines multiple radial models separately along some number of
    rays projected from the center. These rays smoothly transition from one to
    another along angles theta. The ray transition uses a cosine smoothing
    function which depends on the number of rays, for example with two rays the
    brightness would be:

    $$I(R,theta) = I_1(R)*\\cos(\\theta \\% \\pi) + I_2(R)*\\cos((theta + \\pi/2) \\% \\pi)$$

    For `theta = 0` the brightness comes entirely from `I_1` while for `theta = pi/2`
    the brightness comes entirely from `I_2`.

    Options:
        symmetric: If True, the model will have symmetry for rotations of pi radians
            and each ray will appear twice on the sky on opposite sides of the model.
            If False, each ray is independent.
        segments: The number of segments to divide the model into. This controls
            how many rays are used in the model. The default is 2
    """

    _model_type = "ray"
    _options = ("symmetric", "segments")

    def __init__(self, *args, symmetric=True, segments=2, **kwargs):
        super().__init__(*args, **kwargs)
        self.symmetric = symmetric
        self.segments = segments

    def polar_model(self, R, T):
        model = torch.zeros_like(R)
        weight = torch.zeros_like(R)
        cycle = np.pi if self.symmetric else 2 * np.pi
        w = cycle / self.segments
        v = w * np.arange(self.segments)
        for s in range(self.segments):
            angles = (T + cycle / 2 - v[s]) % cycle - cycle / 2
            indices = (angles >= -w) & (angles < w)
            weights = (torch.cos(angles[indices] * self.segments) + 1) / 2
            model[indices] += weights * self.iradial_model(s, R[indices])
            weight[indices] += weights
        return model / weight

    def brightness(self, x, y):
        x, y = self.transform_coordinates(x, y)
        return self.polar_model(self.radius_metric(x, y), self.angular_metric(x, y))



from typing import Union

import torch
import numpy as np
from astropy.io import fits

from ...utils.initialize import auto_variance
from ... import config
from ...errors import SpecificationConflict
from ..image_object import Image
from ..window import Window


class DataMixin:

    def __init__(
        self,
        *args,
        mask=None,
        std=None,
        variance=None,
        weight=None,
        _mask=None,
        _weight=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if _mask is None:
            self.mask = mask
        else:
            self._mask = _mask
        if (std is not None) + (variance is not None) + (weight is not None) > 1:
            raise SpecificationConflict(
                "Can only define one of: std, variance, or weight for a given image."
            )

        if _weight is not None:
            self._weight = _weight
        elif std is not None:
            self.std = std
        elif variance is not None:
            self.variance = variance
        else:
            self.weight = weight

        # Set nan pixels to be masked automatically
        if torch.any(torch.isnan(self.data)).item():
            self._mask = self.mask | torch.isnan(self.data)

    @property
    def std(self):
        """Stores the standard deviation of the image pixels. This represents
        the uncertainty in each pixel value. It should always have the
        same shape as the image data. In the case where the standard
        deviation is not known, a tensor of ones will be created to
        stand in as the standard deviation values.

        The standard deviation is not stored directly, instead it is
        computed as :math:`\\sqrt{1/W}` where :math:`W` is the
        weights.

        """
        if self.has_variance:
            return torch.sqrt(self.variance)
        return torch.ones_like(self.data)

    @std.setter
    def std(self, std):
        if std is None:
            self._weight = None
            return
        if isinstance(std, str) and std == "auto":
            self.weight = "auto"
            return
        self.weight = 1 / std**2

    @property
    def has_std(self):
        """Returns True when the image object has stored standard deviation values. If
        this is False and the std property is called then a
        tensor of ones will be returned.

        """
        try:
            return self._weight is not None
        except AttributeError:
            return False

    @property
    def variance(self):
        """Stores the variance of the image pixels. This represents the
        uncertainty in each pixel value. It should always have the
        same shape as the image data. In the case where the variance
        is not known, a tensor of ones will be created to stand in as
        the variance values.

        The variance is not stored directly, instead it is
        computed as :math:`\\frac{1}{W}` where :math:`W` is the
        weights.

        """
        if self.has_variance:
            return torch.where(self._weight == 0, torch.inf, 1 / self._weight)
        return torch.ones_like(self.data)

    @variance.setter
    def variance(self, variance):
        if variance is None:
            self._weight = None
            return
        if isinstance(variance, str) and variance == "auto":
            self.weight = "auto"
            return
        self.weight = 1 / variance

    @property
    def has_variance(self):
        """Returns True when the image object has stored variance values. If
        this is False and the variance property is called then a
        tensor of ones will be returned.

        """
        try:
            return self._weight is not None
        except AttributeError:
            return False

    @property
    def weight(self):
        """Stores the weight of the image pixels. This represents the
        uncertainty in each pixel value. It should always have the
        same shape as the image data. In the case where the weight
        is not known, a tensor of ones will be created to stand in as
        the weight values.

        The weights are used to proprtionately scale residuals in the
        likelihood. Most commonly this shows up as a :math:`\\chi^2`
        like:

        .. math::

          \\chi^2 = (\\vec{y} - \\vec{f(\\theta)})^TW(\\vec{y} - \\vec{f(\\theta)})

        which can be optimized to find parameter values. Using the
        Jacobian, which in this case is the derivative of every pixel
        wrt every parameter, the weight matrix also appears in the
        gradient:

        .. math::

          \\vec{g} = J^TW(\\vec{y} - \\vec{f(\\theta)})

        and the hessian approximation used in Levenberg-Marquardt:

        .. math::

          H \\approx J^TWJ

        """
        if self.has_weight:
            return self._weight
        return torch.ones_like(self.data)

    @weight.setter
    def weight(self, weight):
        if weight is None:
            self._weight = None
            return
        if isinstance(weight, str) and weight == "auto":
            weight = 1 / auto_variance(self.data, self.mask).T
        self._weight = torch.transpose(
            torch.as_tensor(weight, dtype=config.DTYPE, device=config.DEVICE), 0, 1
        )
        if self._weight.shape != self.data.shape:
            self._weight = None
            raise SpecificationConflict(
                f"weight/variance must have same shape as data ({weight.shape} vs {self.data.shape})"
            )

    @property
    def has_weight(self):
        """Returns True when the image object has stored weight values. If
        this is False and the weight property is called then a
        tensor of ones will be returned.

        """
        try:
            return self._weight is not None
        except AttributeError:
            self._weight = None
            return False

    @property
    def mask(self):
        """The mask stores a tensor of boolean values which indicate any
        pixels to be ignored. These pixels will be skipped in
        likelihood evaluations and in parameter optimization. It is
        common practice to mask pixels with pathological values such
        as due to cosmic rays or satellites passing through the image.

        In a mask, a True value indicates that the pixel is masked and
        should be ignored. False indicates a normal pixel which will
        inter into most calculations.

        If no mask is provided, all pixels are assumed valid.

        """
        if self.has_mask:
            return self._mask
        return torch.zeros_like(self.data, dtype=torch.bool)

    @mask.setter
    def mask(self, mask):
        if mask is None:
            self._mask = None
            return
        self._mask = torch.transpose(
            torch.as_tensor(mask, dtype=torch.bool, device=config.DEVICE), 0, 1
        )
        if self._mask.shape != self.data.shape:
            self._mask = None
            raise SpecificationConflict(
                f"mask must have same shape as data ({mask.shape} vs {self.data.shape})"
            )

    @property
    def has_mask(self):
        """
        Single boolean to indicate if a mask has been provided by the user.
        """
        try:
            return self._mask is not None
        except AttributeError:
            return False

    def to(self, dtype=None, device=None):
        """Converts the stored `Target_Image` data, variance, psf, etc to a
        given data type and device.

        """
        if dtype is not None:
            dtype = config.DTYPE
        if device is not None:
            device = config.DEVICE
        super().to(dtype=dtype, device=device)

        if self.has_weight:
            self._weight = self._weight.to(dtype=dtype, device=device)
        if self.has_mask:
            self._mask = self._mask.to(dtype=torch.bool, device=device)
        return self

    def copy(self, **kwargs):
        """Produce a copy of this image with all of the same properties. This
        can be used when one wishes to make temporary modifications to
        an image and then will want the original again.

        """
        kwargs = {"_mask": self._mask, "_weight": self._weight, **kwargs}
        return super().copy(**kwargs)

    def blank_copy(self, **kwargs):
        """Produces a blank copy of the image which has the same properties
        except that its data is now filled with zeros.

        """
        kwargs = {"_mask": self._mask, "_weight": self._weight, **kwargs}
        return super().blank_copy(**kwargs)

    def get_window(self, other: Union[Image, Window], indices=None, **kwargs):
        """Get a sub-region of the image as defined by an other image on the sky."""
        if indices is None:
            indices = self.get_indices(other if isinstance(other, Window) else other.window)
        return super().get_window(
            other,
            _weight=self._weight[indices] if self.has_weight else None,
            _mask=self._mask[indices] if self.has_mask else None,
            indices=indices,
            **kwargs,
        )

    def fits_images(self):
        images = super().fits_images()
        if self.has_weight:
            images.append(
                fits.ImageHDU(
                    torch.transpose(self.weight, 0, 1).detach().cpu().numpy(), name="WEIGHT"
                )
            )
        if self.has_mask:
            images.append(
                fits.ImageHDU(
                    torch.transpose(self.mask, 0, 1).detach().cpu().numpy().astype(int), name="MASK"
                )
            )
        return images

    def load(self, filename: str, hduext=0):
        """Load the image from a FITS file. This will load the data, WCS, and
        any ancillary data such as variance, mask, and PSF.

        """
        hdulist = super().load(filename, hduext=hduext)
        if "WEIGHT" in hdulist:
            self.weight = np.array(hdulist["WEIGHT"].data, dtype=np.float64)
        if "MASK" in hdulist:
            self.mask = np.array(hdulist["MASK"].data, dtype=bool)
        elif "DQ" in hdulist:
            self.mask = np.array(hdulist["DQ"].data, dtype=bool)
        return hdulist

    def reduce(self, scale, **kwargs):
        """Returns a new `Target_Image` object with a reduced resolution
        compared to the current image. `scale` should be an integer
        indicating how much to reduce the resolution. If the
        `Target_Image` was originally (48,48) pixels across with a
        pixelscale of 1 and `reduce(2)` is called then the image will
        be (24,24) pixels and the pixelscale will be 2. If `reduce(3)`
        is called then the returned image will be (16,16) pixels
        across and the pixelscale will be 3.

        """
        MS = self.data.shape[0] // scale
        NS = self.data.shape[1] // scale

        return super().reduce(
            scale=scale,
            _weight=(
                1
                / self.variance[: MS * scale, : NS * scale]
                .reshape(MS, scale, NS, scale)
                .sum(axis=(1, 3))
                if self.has_variance
                else None
            ),
            _mask=(
                self.mask[: MS * scale, : NS * scale]
                .reshape(MS, scale, NS, scale)
                .amax(axis=(1, 3))
                if self.has_mask
                else None
            ),
            **kwargs,
        )



from . import run_from_terminal

if __name__ == "__main__":
    """
    Entry point for the CLI (Command Line Interface).

    Calls the run_from_terminal function
    from the __init__.py file
    which is used for handling command line arguments.
    """
    run_from_terminal()



from . import (
    conversions,
    initialize,
    decorators,
    integration,
    interpolate,
    optimization,
    parametric_profiles,
)

__all__ = [
    "optimization",
    "decorators",
    "interpolate",
    "integration",
    "parametric_profiles",
    "initialize",
    "conversions",
]



#!/usr/bin/env python
# coding: utf-8

# # Aligning Images
#
# In AstroPhot, the image WCS is part of the model and so can be optimized alongside other model parameters. Here we will demonstrate a basic example of image alignment, but the sky is the limit, you can perform highly detailed image alignment with AstroPhot!

# In[ ]:


import astrophot as ap
import matplotlib.pyplot as plt
import numpy as np
import torch
import socket

socket.setdefaulttimeout(60)


# ## Relative shift
#
# Often the WCS solution is already really good, we just need a local shift in x and/or y to get things just right. Lets start by optimizing a translation in the WCS that improves the fit for our models!

# In[ ]:


target_r = ap.TargetImage(
    filename="https://www.legacysurvey.org/viewer/fits-cutout?ra=329.2715&dec=13.6483&size=150&layer=ls-dr9&pixscale=0.262&bands=r",
    name="target_r",
    variance="auto",
)
target_g = ap.TargetImage(
    filename="https://www.legacysurvey.org/viewer/fits-cutout?ra=329.2715&dec=13.6483&size=150&layer=ls-dr9&pixscale=0.262&bands=g",
    name="target_g",
    variance="auto",
)

# Uh-oh! our images are misaligned by 1 pixel, this will cause problems!
target_g.crpix = target_g.crpix + 1

fig, axarr = plt.subplots(1, 2, figsize=(15, 7))
ap.plots.target_image(fig, axarr[0], target_r)
axarr[0].set_title("Target Image (r-band)")
ap.plots.target_image(fig, axarr[1], target_g)
axarr[1].set_title("Target Image (g-band)")
plt.show()


# In[ ]:


# fmt: off
# r-band model
psfr = ap.Model(name="psfr", model_type="moffat psf model", n=2, Rd=1.0, target=target_r.psf_image(data=np.zeros((51, 51))))
star1r = ap.Model(name="star1-r", model_type="point model", window=[0, 60, 80, 135], center=[12, 9], psf=psfr, target=target_r)
star2r = ap.Model(name="star2-r", model_type="point model", window=[40, 90, 20, 70], center=[3, -7], psf=psfr, target=target_r)
star3r = ap.Model(name="star3-r", model_type="point model", window=[109, 150, 40, 90], center=[-15, -3], psf=psfr, target=target_r)
modelr = ap.Model(name="model-r", model_type="group model", models=[star1r, star2r, star3r], target=target_r)

# g-band model
psfg = ap.Model(name="psfg", model_type="moffat psf model", n=2, Rd=1.0, target=target_g.psf_image(data=np.zeros((51, 51))))
star1g = ap.Model(name="star1-g", model_type="point model", window=[0, 60, 80, 135], center=star1r.center, psf=psfg, target=target_g)
star2g = ap.Model(name="star2-g", model_type="point model", window=[40, 90, 20, 70], center=star2r.center, psf=psfg, target=target_g)
star3g = ap.Model(name="star3-g", model_type="point model", window=[109, 150, 40, 90], center=star3r.center, psf=psfg, target=target_g)
modelg = ap.Model(name="model-g", model_type="group model", models=[star1g, star2g, star3g], target=target_g)

# total model
target_full = ap.TargetImageList([target_r, target_g])
model = ap.Model(name="model", model_type="group model", models=[modelr, modelg], target=target_full)

# fmt: on
fig, axarr = plt.subplots(1, 2, figsize=(15, 7))
ap.plots.target_image(fig, axarr, target_full)
axarr[0].set_title("Target Image (r-band)")
axarr[1].set_title("Target Image (g-band)")
ap.plots.model_window(fig, axarr[0], modelr)
ap.plots.model_window(fig, axarr[1], modelg)
plt.show()


# In[ ]:


model.initialize()
res = ap.fit.LM(model, verbose=1).fit()
fig, axarr = plt.subplots(2, 2, figsize=(15, 10))
ap.plots.model_image(fig, axarr[0], model)
axarr[0, 0].set_title("Model Image (r-band)")
axarr[0, 1].set_title("Model Image (g-band)")
ap.plots.residual_image(fig, axarr[1], model)
axarr[1, 0].set_title("Residual Image (r-band)")
axarr[1, 1].set_title("Residual Image (g-band)")
plt.show()


# Here we see a clear signal of an image misalignment, in the g-band all of the residuals have a dipole in the same direction! Lets free up the position of the g-band image and optimize a shift. This only requires a single line of code!

# In[ ]:


target_g.crtan.to_dynamic()


# Now we can optimize the model again, notice how it now has two more parameters. These are the x,y position of the image in the tangent plane. See the AstroPhot coordinate description on the website for more details on why this works.

# In[ ]:


res = ap.fit.LM(model, verbose=1).fit()
fig, axarr = plt.subplots(2, 2, figsize=(15, 10))
ap.plots.model_image(fig, axarr[0], model)
axarr[0, 0].set_title("Model Image (r-band)")
axarr[0, 1].set_title("Model Image (g-band)")
ap.plots.residual_image(fig, axarr[1], model)
axarr[1, 0].set_title("Residual Image (r-band)")
axarr[1, 1].set_title("Residual Image (g-band)")
plt.show()


# Yay! no more dipole. The fits aren't the best, clearly these objects aren't super well described by a single moffat model. But the main goal today was to show that we could align the images very easily. Note, its probably best to start with a reasonably good WCS from the outset, and this two stage approach where we optimize the models and then optimize the models plus a shift might be more stable than just fitting everything at once from the outset. Often for more complex models it is best to start with a simpler model and fit each time you introduce more complexity.

# ## Shift and rotation
#
# Lets say we really don't trust our WCS, we think something has gone wrong and we want freedom to fully shift and rotate the relative positions of the images relative to each other. How can we do this?

# In[ ]:


def rotate(phi):
    """Create a 2D rotation matrix for a given angle in radians."""
    return torch.stack(
        [
            torch.stack([torch.cos(phi), -torch.sin(phi)]),
            torch.stack([torch.sin(phi), torch.cos(phi)]),
        ]
    )


# Uh-oh! Our image is misaligned by some small angle
target_g.CD = target_g.CD.value @ rotate(torch.tensor(np.pi / 32, dtype=torch.float64))
# Uh-oh! our alignment from before has been erased
target_g.crtan.value = (0, 0)


# In[ ]:


fig, axarr = plt.subplots(2, 2, figsize=(15, 10))
ap.plots.model_image(fig, axarr[0], model)
axarr[0, 0].set_title("Model Image (r-band)")
axarr[0, 1].set_title("Model Image (g-band)")
ap.plots.residual_image(fig, axarr[1], model)
axarr[1, 0].set_title("Residual Image (r-band)")
axarr[1, 1].set_title("Residual Image (g-band)")
plt.show()


# Notice that there is not a universal dipole like in the shift example. Most of the offset is caused by the rotation in this example.

# In[ ]:


# this will control the relative rotation of the g-band image
phi = ap.Param(name="phi", dynamic_value=0.0, dtype=torch.float64)

# Set the target_g CD matrix to be a function of the rotation angle
# The CD matrix can encode rotation, skew, and rectangular pixels. We
# are only interested in the rotation here.
init_CD = target_g.CD.value.clone()
target_g.CD = lambda p: init_CD @ rotate(p.phi.value)
target_g.CD.link(phi)

# also optimize the shift of the g-band image
target_g.crtan.to_dynamic()


# In[ ]:


res = ap.fit.LM(model, verbose=1).fit()
fig, axarr = plt.subplots(2, 2, figsize=(15, 10))
ap.plots.model_image(fig, axarr[0], model)
axarr[0, 0].set_title("Model Image (r-band)")
axarr[0, 1].set_title("Model Image (g-band)")
ap.plots.residual_image(fig, axarr[1], model)
axarr[1, 0].set_title("Residual Image (r-band)")
axarr[1, 1].set_title("Residual Image (g-band)")
plt.show()


# In[ ]:



from matplotlib.pyplot import get_cmap

# from matplotlib.colors import ListedColormap
# import numpy as np

__all__ = ["main_pallet", "cmap_grad", "cmap_div"]

main_pallet = {
    "primary1": "tab:blue",
    "primary2": "tab:orange",
    "primary3": "tab:red",
    "secondary1": "tab:green",
    "secondary2": "tab:purple",
    "pop": "tab:pink",
}

cmap_grad = get_cmap("inferno")
cmap_div = get_cmap("seismic")  # twilight  RdBu_r
# print(__file__)
# colors = np.load(f"{__file__[:-10]}/managua_cmap.npy")
# cmap_div = ListedColormap(list(reversed(colors)), name="mangua")



import numpy as np
import astrophot as ap
from astropy.wcs import WCS


def get_astropy_wcs():
    hdr = {
        "SIMPLE": "T",
        "NAXIS": 2,
        "NAXIS1": 180,
        "NAXIS2": 170,
        "CTYPE1": "RA---TAN",
        "CTYPE2": "DEC--TAN",
        "CRVAL1": 195.0588,
        "CRVAL2": 28.0608,
        "CRPIX1": 90.5,
        "CRPIX2": 85.5,
        "CD1_1": -0.000416666666666667,
        "CD1_2": 0.0,
        "CD2_1": 0.0,
        "CD2_2": 0.000416666666666667,
        # "IMAGEW": 180.0,
        # "IMAGEH": 170.0,
    }
    return WCS(hdr)


def make_basic_sersic(
    N=52,
    M=50,
    pixelscale=0.8,
    x=20.5,
    y=21.4,
    PA=45 * np.pi / 180,
    q=0.7,
    n=1.5,
    Re=15.1,
    Ie=10.0,
    rand=12345,
    **kwargs,
):

    np.random.seed(rand)
    mask = np.zeros((N, M), dtype=bool)
    mask[0][0] = True
    target = ap.TargetImage(
        data=np.zeros((N, M)),
        pixelscale=pixelscale,
        psf=ap.utils.initialize.gaussian_psf(2 / pixelscale, 11, pixelscale),
        mask=mask,
        **kwargs,
    )

    MODEL = ap.models.SersicGalaxy(
        name="basic sersic model",
        target=target,
        center=[x, y],
        PA=PA,
        q=q,
        n=n,
        Re=Re,
        Ie=Ie,
        sampling_mode="quad:5",
    )

    img = MODEL().data.T.detach().cpu().numpy()
    target.data = (
        img
        + np.random.normal(scale=0.1, size=img.shape)
        + np.random.normal(scale=np.sqrt(img) / 10)
    )
    target.variance = 0.1**2 + img / 100

    return target


def make_basic_gaussian(
    N=50,
    M=50,
    pixelscale=0.8,
    x=24.5,
    y=25.4,
    PA=45 * np.pi / 180,
    sigma=3,
    flux=1,
    rand=12345,
):

    np.random.seed(rand)
    target = ap.TargetImage(
        data=np.zeros((N, M)),
        pixelscale=pixelscale,
        psf=ap.utils.initialize.gaussian_psf(2 / pixelscale, 11, pixelscale),
    )

    MODEL = ap.models.GaussianGalaxy(
        name="basic gaussian source",
        target=target,
        center=[x, y],
        sigma=sigma,
        flux=flux,
        PA=0.0,
        q=0.99,
    )

    img = MODEL().data.detach().cpu().numpy()
    target.data = (
        img
        + np.random.normal(scale=0.1, size=img.shape)
        + np.random.normal(scale=np.sqrt(img) / 10)
    )
    target.variance = 0.1**2 + img / 100

    return target


def make_basic_gaussian_psf(
    N=25,
    pixelscale=0.8,
    sigma=4,
    rand=12345,
):

    np.random.seed(rand)
    psf = ap.utils.initialize.gaussian_psf(sigma * pixelscale, N, pixelscale)
    target = ap.PSFImage(
        data=psf + np.random.normal(scale=np.sqrt(psf) / 20),
        pixelscale=pixelscale,
        variance=psf / 400,
    )
    target.normalize()

    return target



# Metropolis-Hasting Markov-Chain Monte-Carlo
from typing import Optional, Sequence

import torch
import numpy as np

try:
    import emcee
except ImportError:
    emcee = None

from .base import BaseOptimizer
from ..models import Model
from .. import config

__all__ = ["MHMCMC"]


class MHMCMC(BaseOptimizer):
    """Metropolis-Hastings Markov-Chain Monte-Carlo sampler, based on:
    https://en.wikipedia.org/wiki/Metropolis-Hastings_algorithm . This is simply
    a thin wrapper for the Emcee package, which is a well-known MCMC sampler.
    """

    def __init__(
        self,
        model: Model,
        initial_state: Optional[Sequence] = None,
        max_iter: int = 1000,
        likelihood="gaussian",
        **kwargs,
    ):
        super().__init__(model, initial_state, max_iter=max_iter, **kwargs)

        if emcee is None:
            raise ImportError(
                "The emcee package is required for MHMCMC sampling. Please install it with `pip install emcee` or the like."
            )
        self.likelihood = likelihood

        self.chain = []

    def density(self, state: np.ndarray) -> np.ndarray:
        """
        Returns the density of the model at the given state vector.
        This is used to calculate the likelihood of the model at the given state.
        """
        state = torch.tensor(state, dtype=config.DTYPE, device=config.DEVICE)
        if self.likelihood == "gaussian":
            return np.array(list(self.model.gaussian_log_likelihood(s).item() for s in state))
        elif self.likelihood == "poisson":
            return np.array(list(self.model.poisson_log_likelihood(s).item() for s in state))
        else:
            raise ValueError(f"Unknown likelihood type: {self.likelihood}")

    def fit(
        self,
        state: Optional[np.ndarray] = None,
        nsamples: Optional[int] = None,
        restart_chain: bool = True,
        skip_initial_state_check: bool = True,
        flat_chain: bool = True,
    ):
        """
        Performs the MCMC sampling using a Metropolis Hastings acceptance step and records the chain for later examination.
        """

        if nsamples is None:
            nsamples = self.max_iter

        if state is None:
            state = self.current_state

        if len(state.shape) == 1:
            nwalkers = state.shape[0] * 2
            state = state * np.random.normal(loc=1, scale=0.01, size=(nwalkers, state.shape[0]))
        else:
            nwalkers = state.shape[0]
        ndim = state.shape[1]
        sampler = emcee.EnsembleSampler(nwalkers, ndim, self.density, vectorize=True)
        state = sampler.run_mcmc(state, nsamples, skip_initial_state_check=skip_initial_state_check)
        if restart_chain:
            self.chain = sampler.get_chain(flat=flat_chain)
        else:
            self.chain = np.append(self.chain, sampler.get_chain(flat=flat_chain), axis=0)
        self.model.fill_dynamic_values(
            torch.tensor(self.chain[-1], dtype=config.DTYPE, device=config.DEVICE)
        )
        return self



import torch
import numpy as np

from ...param import forward
from ...utils.decorators import ignore_numpy_warnings
from .._shared_methods import parametric_initialize, parametric_segment_initialize
from ...utils.parametric_profiles import moffat_np
from .. import func


def _x0_func(model_params, R, F):
    return 2.0, R[4], 10 ** F[0]


class MoffatMixin:
    """Moffat radial light profile (Moffat 1969).

    The moffat profile gives a good representation of the gneeral structure of
    PSF functions for ground based data. It can also be used to fit extended
    objects. The functional form of the Moffat profile is defined as:

    $$I(R) = \\frac{I_0}{(1 + (R/R_d)^2)^n}$$

    n is the concentration index which controls the shape of the profile.

    Parameters:
        n: Concentration index which controls the shape of the brightness profile
        Rd: Scale length radius
        I0: Intensity at the center of the profile
    """

    _model_type = "moffat"
    _parameter_specs = {
        "n": {"units": "none", "valid": (0.1, 10), "shape": ()},
        "Rd": {"units": "arcsec", "valid": (0, None), "shape": ()},
        "I0": {"units": "flux/arcsec^2", "valid": (0, None), "shape": ()},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_initialize(
            self,
            self.target[self.window],
            moffat_np,
            ("n", "Rd", "I0"),
            _x0_func,
        )

    @forward
    def radial_model(self, R, n, Rd, I0):
        return func.moffat(R, n, Rd, I0)


class iMoffatMixin:
    """Moffat radial light profile (Moffat 1969).

    The moffat profile gives a good representation of the gneeral structure of
    PSF functions for ground based data. It can also be used to fit extended
    objects. The functional form of the Moffat profile is defined as:

    $$I(R) = \\frac{I_0}{(1 + (R/R_d)^2)^n}$$

    n is the concentration index which controls the shape of the profile.

    `n`, `Rd`, and `I0` are batched by their first dimension, allowing for
    multiple Moffat profiles to be defined at once.

    Parameters:
        n: Concentration index which controls the shape of the brightness profile
        Rd: Scale length radius
        I0: Intensity at the center of the profile
    """

    _model_type = "moffat"
    _parameter_specs = {
        "n": {"units": "none", "valid": (0.1, 10)},
        "Rd": {"units": "arcsec", "valid": (0, None)},
        "I0": {"units": "flux/arcsec^2", "valid": (0, None)},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_segment_initialize(
            model=self,
            target=self.target[self.window],
            prof_func=moffat_np,
            params=("n", "Rd", "I0"),
            x0_func=_x0_func,
            segments=self.segments,
        )

    @forward
    def iradial_model(self, i, R, n, Rd, I0):
        return func.moffat(R, n[i], Rd[i], I0[i])



import numpy as np
import torch

from ...utils.decorators import ignore_numpy_warnings
from ...utils.interpolate import default_prof
from ...param import forward
from .. import func
from ... import config


class InclinedMixin:
    """A model which defines a position angle and axis ratio.

    PA and q operate on the coordinates to transform the model. Given some x,y
    the updated values are:

    $$x', y' = \\rm{rotate}(-PA + \\pi/2, x, y)$$
    $$y'' = y' / q$$

    where x' and y'' are the final transformed coordinates. The pi/2 is included
    such that the position angle is defined with 0 at north. The -PA is such
    that the position angle increases to the East. Thus, the position angle is a
    standard East of North definition assuming the WCS of the image is correct.

    Note that this means radii are defined with $R = \\sqrt{x^2 +
    (\\frac{y}{q})^2}$ rather than the common alternative which is $R =
    \\sqrt{qx^2 + \\frac{y^2}{q}}$
    """

    _parameter_specs = {
        "q": {"units": "b/a", "valid": (0.01, 1), "shape": ()},
        "PA": {"units": "radians", "valid": (0, np.pi), "cyclic": True, "shape": ()},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if self.PA.initialized and self.q.initialized:
            return
        target_area = self.target[self.window]
        dat = target_area.data.detach().cpu().numpy().copy()
        if target_area.has_mask:
            mask = target_area.mask.detach().cpu().numpy()
            dat[mask] = np.median(dat[~mask])
        edge = np.concatenate((dat[:, 0], dat[:, -1], dat[0, :], dat[-1, :]))
        edge_average = np.nanmedian(edge)
        dat -= edge_average
        x, y = target_area.coordinate_center_meshgrid()
        x = (x - self.center.value[0]).detach().cpu().numpy()
        y = (y - self.center.value[1]).detach().cpu().numpy()
        mu20 = np.median(dat * np.abs(x))
        mu02 = np.median(dat * np.abs(y))
        mu11 = np.median(dat * x * y / np.sqrt(np.abs(x * y) + self.softening**2))
        M = np.array([[mu20, mu11], [mu11, mu02]])
        if not self.PA.initialized:
            if np.any(np.iscomplex(M)) or np.any(~np.isfinite(M)):
                self.PA.dynamic_value = np.pi / 2
            else:
                self.PA.dynamic_value = (
                    0.5 * np.arctan2(2 * mu11, mu20 - mu02) - np.pi / 2
                ) % np.pi
        if not self.q.initialized:
            if np.any(np.iscomplex(M)) or np.any(~np.isfinite(M)):
                l = (0.7, 1.0)
            else:
                l = np.sort(np.linalg.eigvals(M))
            self.q.dynamic_value = np.clip(np.sqrt(np.abs(l[0] / l[1])), 0.1, 0.9)

    @forward
    def transform_coordinates(self, x, y, PA, q):
        x, y = super().transform_coordinates(x, y)
        x, y = func.rotate(-PA + np.pi / 2, x, y)
        return x, y / q


class SuperEllipseMixin:
    """Generalizes the definition of radius and so modifies the evaluation of radial models.

    A superellipse transformation allows for the expression of "boxy" and
    "disky" modifications to traditional elliptical isophotes. This is a common
    extension of the standard elliptical representation, especially
    for early-type galaxies. The functional form for this is:

    $$R = (|x|^C + |y|^C)^(1/C)$$

    where R is the new distance metric, X Y are the coordinates, and C
    is the coefficient for the superellipse. C can take on any value
    greater than zero where C = 2 is the standard distance metric, 0 <
    C < 2 creates disky or pointed perturbations to an ellipse, and C
    > 2 transforms an ellipse to be more boxy.

    Parameters:
        C: superellipse distance metric parameter.

    """

    _model_type = "superellipse"
    _parameter_specs = {
        "C": {"units": "none", "dynamic_value": 2.0, "valid": (0, None)},
    }

    @forward
    def radius_metric(self, x, y, C):
        return torch.pow(x.abs().pow(C) + y.abs().pow(C) + self.softening**C, 1.0 / C)


class FourierEllipseMixin:
    """Sine wave perturbation of the elliptical radius metric.

    This allows for the expression of arbitrarily complex isophotes instead of
    pure ellipses. This is a common extension of the standard elliptical
    representation. The form of the Fourier perturbations is:

    $$R' = R * \\exp(\\sum_m(a_m * \\cos(m * \\theta + \\phi_m)))$$

    where R' is the new radius value, R is the original radius (typically
    computed as $\\sqrt{x^2+y^2}$), m is the index of the Fourier mode, a_m is
    the amplitude of the m'th Fourier mode, theta is the angle around the
    ellipse (typically $\\arctan(y/x)$), and phi_m is the phase of the m'th
    fourier mode.

    One can create extremely complex shapes using different Fourier modes,
    however usually it is only low order modes that are of interest. For
    intuition, the first Fourier mode is roughly equivalent to a lopsided
    galaxy, one side will be compressed and the opposite side will be expanded.
    The second mode is almost never used as it is nearly degenerate with
    ellipticity. The third mode is an alternate kind of lopsidedness for a
    galaxy which makes it somewhat triangular, meaning that it is wider on one
    side than the other. The fourth mode is similar to a boxyness/diskyness
    parameter of a superelllipse which tends to make more pronounced peanut
    shapes since it is more rounded than a superellipse representation. Modes
    higher than 4 are only useful in very specialized situations. In general one
    should consider carefully why the Fourier modes are being used for the
    science case at hand.

    Parameters:
        am: Tensor of amplitudes for the Fourier modes, indicates the strength
            of each mode.
        phim: Tensor of phases for the Fourier modes, adjusts the
            orientation of the mode perturbation relative to the major axis. It
            is cyclically defined in the range [0,2pi)

    Options:
        modes: Tuple of integers indicating which Fourier modes to use.
    """

    _model_type = "fourier"
    _parameter_specs = {
        "am": {"units": "none"},
        "phim": {"units": "radians", "valid": (0, 2 * np.pi), "cyclic": True},
    }
    _options = ("modes",)

    def __init__(self, *args, modes=(3, 4), **kwargs):
        super().__init__(*args, **kwargs)
        self.modes = torch.tensor(modes, dtype=config.DTYPE, device=config.DEVICE)

    @forward
    def radius_metric(self, x, y, am, phim):
        R = super().radius_metric(x, y)
        theta = self.angular_metric(x, y)
        return R * torch.exp(
            torch.sum(
                am.unsqueeze(-1)
                * torch.cos(self.modes.unsqueeze(-1) * theta.flatten() + phim.unsqueeze(-1)),
                0,
            ).reshape(x.shape)
        )

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if not self.am.initialized:
            self.am.dynamic_value = np.zeros(len(self.modes))
        if not self.phim.initialized:
            self.phim.value = np.zeros(len(self.modes))


class WarpMixin:
    """Warped model with varying PA and q as a function of radius.

    This works by warping the coordinates using the same transform for a global
    PA, q except applied to each pixel individually based on its unwarped radius
    value. In the limit that PA and q are a constant, this recovers a basic
    model with global PA, q. However, a linear PA profile will give a spiral
    appearance, variations of PA, q profiles can create complex galaxy models.
    The form of the coordinate transformation for each pixel looks like:

    $$R = \\sqrt{x^2 + y^2}$$
    $$x', y' = \\rm{rotate}(-PA(R) + \\pi/2, x, y)$$
    $$y'' = y' / q(R)$$

    Note that now PA and q are functions of radius R, which is computed from the
    original coordinates X, Y. This is achieved by making PA and q a spline
    profile.

    Parameters:
        q_R: Tensor of axis ratio values for axis ratio spline
        PA_R: Tensor of position angle values as input to the spline

    """

    _model_type = "warp"
    _parameter_specs = {
        "q_R": {"units": "b/a", "valid": (0, 1)},
        "PA_R": {"units": "radians", "valid": (0, np.pi), "cyclic": True},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if not self.PA_R.initialized:
            if self.PA_R.prof is None:
                self.PA_R.prof = default_prof(self.window.shape, self.target.pixelscale, 2, 0.2)
            self.PA_R.dynamic_value = np.zeros(len(self.PA_R.prof)) + np.pi / 2
        if not self.q_R.initialized:
            if self.q_R.prof is None:
                self.q_R.prof = default_prof(self.window.shape, self.target.pixelscale, 2, 0.2)
            self.q_R.dynamic_value = np.ones(len(self.q_R.prof)) * 0.8

    @forward
    def transform_coordinates(self, x, y, q_R, PA_R):
        x, y = super().transform_coordinates(x, y)
        R = self.radius_metric(x, y)
        PA = func.spline(R, self.PA_R.prof, PA_R, extend="const")
        q = func.spline(R, self.q_R.prof, q_R, extend="const")
        x, y = func.rotate(-PA + np.pi / 2, x, y)
        return x, y / q


class TruncationMixin:
    """Truncated model with radial brightness profile.

    This model will smoothly truncate the radial brightness profile at Rt. The
    truncation is centered on Rt and thus two identical models with the same Rt
    (and St) where one is inner truncated and the other is outer truncated will
    reproduce nearly the same as a single un-truncated model.

    By default the St parameter is set fixed to 1.0, giving a relatively smooth
    truncation. This can be set to a smaller value for sharper truncations or a
    larger value for even more gradual truncation. It can be set dynamic to be
    optimized in a model, though it is possible for this parameter to be
    unstable if there isn't a clear truncation signal in the data.

    Parameters:
        Rt: The truncation radius in arcseconds.
        St: The steepness of the truncation profile, controlling how quickly
            the brightness drops to zero at the truncation radius.

    Options:
        outer_truncation: If True, the model will truncate the brightness beyond
            the truncation radius. If False, the model will truncate the
            brightness within the truncation radius.
    """

    _model_type = "truncated"
    _parameter_specs = {
        "Rt": {"units": "arcsec", "valid": (0, None), "shape": ()},
        "St": {"units": "none", "valid": (0, None), "shape": (), "value": 1.0},
    }
    _options = ("outer_truncation",)

    def __init__(self, *args, outer_truncation=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.outer_truncation = outer_truncation

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()
        if not self.Rt.initialized:
            prof = default_prof(self.window.shape, self.target.pixelscale, 2, 0.2)
            self.Rt.dynamic_value = prof[len(prof) // 2]

    @forward
    def radial_model(self, R, Rt, St):
        I = super().radial_model(R)
        if self.outer_truncation:
            return I * (1 - torch.tanh(St * (R - Rt))) / 2
        return I * (torch.tanh(St * (R - Rt)) + 1) / 2



import numpy as np
import matplotlib.pyplot as plt

import astrophot as ap
from utils import make_basic_sersic, make_basic_gaussian_psf
import pytest


"""
Can't test visuals, so this only tests that the code runs
"""


def test_target_image():
    target = make_basic_sersic()
    try:
        fig, ax = plt.subplots()
    except Exception:
        pytest.skip("skipping test_target_image because matplotlib is not installed properly")
    ap.plots.target_image(fig, ax, target)
    plt.close()


def test_psf_image():
    target = make_basic_gaussian_psf()
    try:
        fig, ax = plt.subplots()
    except Exception:
        pytest.skip("skipping test_target_image because matplotlib is not installed properly")
    ap.plots.psf_image(fig, ax, target)
    plt.close()


def test_target_image_list():
    target1 = make_basic_sersic(name="target1")
    target2 = make_basic_sersic(name="target2")
    target = ap.TargetImageList([target1, target2])
    try:
        fig, ax = plt.subplots(2)
    except Exception:
        pytest.skip("skipping test_target_image_list because matplotlib is not installed properly")
    ap.plots.target_image(fig, ax, target)
    plt.close()


def test_model_image():
    target = make_basic_sersic()
    new_model = ap.Model(
        name="constrained sersic",
        model_type="sersic galaxy model",
        center=[20, 20],
        PA=60 * np.pi / 180,
        q=0.5,
        n=2,
        Re=5,
        Ie=1,
        target=target,
    )
    new_model.initialize()
    try:
        fig, ax = plt.subplots()
    except Exception:
        pytest.skip("skipping test because matplotlib is not installed properly")
    ap.plots.model_image(fig, ax, new_model)
    plt.close()


def test_residual_image():
    target = make_basic_sersic()
    new_model = ap.Model(
        name="constrained sersic",
        model_type="sersic galaxy model",
        center=[20, 20],
        PA=60 * np.pi / 180,
        q=0.5,
        n=2,
        Re=5,
        Ie=1,
        target=target,
    )
    new_model.initialize()
    try:
        fig, ax = plt.subplots()
    except Exception:
        pytest.skip("skipping test because matplotlib is not installed properly")
    ap.plots.residual_image(fig, ax, new_model)
    plt.close()


def test_model_windows():
    target = make_basic_sersic()
    new_model = ap.Model(
        name="constrained sersic",
        model_type="sersic galaxy model",
        center=[20, 20],
        PA=60 * np.pi / 180,
        q=0.5,
        n=2,
        Re=5,
        Ie=1,
        window=(10, 10, 30, 30),
        target=target,
    )
    new_model.initialize()
    try:
        fig, ax = plt.subplots()
    except Exception:
        pytest.skip("skipping test because matplotlib is not installed properly")
    ap.plots.model_window(fig, ax, new_model)
    plt.close()


def test_covariance_matrix():
    covariance_matrix = np.array([[1, 0.5], [0.5, 1]])
    mean = np.array([0, 0])
    try:
        fig, ax = plt.subplots()
    except Exception:
        pytest.skip("skipping test because matplotlib is not installed properly")
    fig, ax = ap.plots.covariance_matrix(covariance_matrix, mean, labels=["x", "y"])
    plt.close()


def test_radial_profile():
    target = make_basic_sersic()
    new_model = ap.Model(
        name="constrained sersic",
        model_type="sersic galaxy model",
        center=[20, 20],
        PA=60 * np.pi / 180,
        q=0.5,
        n=2,
        Re=5,
        Ie=1,
        target=target,
    )
    new_model.initialize()
    try:
        fig, ax = plt.subplots()
    except Exception:
        pytest.skip("skipping test because matplotlib is not installed properly")
    ap.plots.radial_light_profile(fig, ax, new_model)
    plt.close()


def test_radial_median_profile():
    target = make_basic_sersic()
    new_model = ap.Model(
        name="constrained sersic",
        model_type="sersic galaxy model",
        center=[20, 20],
        PA=60 * np.pi / 180,
        q=0.5,
        n=2,
        Re=5,
        Ie=1,
        target=target,
    )
    new_model.initialize()
    try:
        fig, ax = plt.subplots()
    except Exception:
        pytest.skip("skipping test because matplotlib is not installed properly")
    ap.plots.radial_median_profile(fig, ax, new_model)
    plt.close()



import unittest
import astrophot as ap
import numpy as np
import torch


######################################################################
# Window List Object
######################################################################


def test_windowlist_creation():

    image1 = ap.Image(
        data=np.zeros((10, 15)),
        pixelscale=1.0,
        zeropoint=1.0,
        name="image1",
    )
    image2 = ap.Image(
        data=np.ones((15, 10)),
        pixelscale=0.5,
        zeropoint=2.0,
        name="image2",
    )
    window1 = ap.Window([4, 13, 5, 9], image1)
    window2 = ap.Window([0, 7, 1, 8], image2)
    windowlist = ap.WindowList([window1, window2])

    window3 = ap.Window([3, 12, 5, 8], image1)
    assert windowlist.index(window3) == 0, "WindowList should find window by index"
    assert len(windowlist) == 2, "WindowList should have two windows"

    window21 = ap.Window([5, 10, 6, 9], image1)
    window22 = ap.Window([0, 9, 0, 8], image2)
    windowlist2 = ap.WindowList([window21, window22])

    windowlist_and = windowlist & windowlist2
    assert len(windowlist_and) == 2, "WindowList should have two windows after intersection"
    assert windowlist_and[0].image is image1, "First window should be from image1"
    assert windowlist_and[1].image is image2, "Second window should be from image2"
    assert windowlist_and[0].i_low == 5, "First window should have i_low of 5"
    assert windowlist_and[0].i_high == 10, "First window should have i_high of 10"
    assert windowlist_and[0].j_low == 6, "First window should have j_low of 6"
    assert windowlist_and[0].j_high == 9, "First window should have j_high of 9"



from .group_model_object import GroupModel
from ..image import PSFImage
from ..errors import InvalidTarget
from ..param import forward

__all__ = ["PSFGroupModel"]


class PSFGroupModel(GroupModel):

    _model_type = "psf"
    usable = True
    normalize_psf = True

    _options = ("normalize_psf",)

    @property
    def target(self):
        try:
            return self._target
        except AttributeError:
            return None

    @target.setter
    def target(self, target):
        if not (target is None or isinstance(target, PSFImage)):
            raise InvalidTarget("Group_Model target must be a PSF_Image instance.")
        try:
            del self._target  # Remove old target if it exists
        except AttributeError:
            pass

        self._target = target

    @forward
    def sample(self, *args, **kwargs):
        """Sample the PSF group model on the target image."""
        psf_img = super().sample(*args, **kwargs)
        if self.normalize_psf:
            psf_img.normalize()
        return psf_img



import astrophot as ap
import torch
import numpy as np

from utils import make_basic_sersic, get_astropy_wcs
import pytest

######################################################################
# Image Objects
######################################################################


@pytest.fixture()
def base_image():
    arr = torch.zeros((10, 15))
    return ap.Image(
        data=arr,
        pixelscale=1.0,
        zeropoint=1.0,
    )


def test_image_creation(base_image):
    base_image.to()
    assert base_image.pixelscale == 1.0, "image should track pixelscale"
    assert base_image.zeropoint == 1.0, "image should track zeropoint"
    assert base_image.crpix[0] == 0, "image should track crpix"
    assert base_image.crpix[1] == 0, "image should track crpix"

    base_image.to(dtype=torch.float64)
    slicer = ap.Window((7, 13, 4, 7), base_image)
    sliced_image = base_image[slicer]
    assert sliced_image.crpix[0] == -7, "crpix of subimage should give relative position"
    assert sliced_image.crpix[1] == -4, "crpix of subimage should give relative position"
    assert sliced_image.shape == (6, 3), "sliced image should have correct shape"


def test_copy(base_image):
    copy_image = base_image.copy()
    assert (
        base_image.pixelscale == copy_image.pixelscale
    ), "copied image should have same pixelscale"
    assert base_image.zeropoint == copy_image.zeropoint, "copied image should have same zeropoint"
    assert (
        base_image.window.extent == copy_image.window.extent
    ), "copied image should have same window"
    copy_image += 1
    assert base_image.data[0][0] == 0.0, "copied image should not share data with original"

    blank_copy_image = base_image.blank_copy()
    assert (
        base_image.pixelscale == blank_copy_image.pixelscale
    ), "copied image should have same pixelscale"
    assert (
        base_image.zeropoint == blank_copy_image.zeropoint
    ), "copied image should have same zeropoint"
    assert (
        base_image.window.extent == blank_copy_image.window.extent
    ), "copied image should have same window"
    blank_copy_image += 1
    assert base_image.data[0][0] == 0.0, "copied image should not share data with original"


def test_image_arithmetic(base_image):
    slicer = ap.Window((-1, 5, 6, 15), base_image)
    sliced_image = base_image[slicer]
    sliced_image += 1

    assert base_image.data[1][8] == 0, "slice should not update base image"
    assert base_image.data[5][5] == 0, "slice should not update base image"

    second_image = ap.Image(
        data=torch.ones((5, 5)),
        pixelscale=1.0,
        zeropoint=1.0,
        crpix=(-1, 1),
    )

    # Test iadd
    base_image += second_image
    assert base_image.data[0][0] == 0, "image addition should only update its region"
    assert base_image.data[3][3] == 1, "image addition should update its region"
    assert base_image.data[3][4] == 0, "image addition should only update its region"
    assert base_image.data[5][3] == 1, "image addition should update its region"

    # Test isubtract
    base_image -= second_image
    assert torch.all(
        torch.isclose(base_image.data, torch.zeros_like(base_image.data))
    ), "image subtraction should only update its region"


def test_image_manipulation():
    new_image = ap.Image(
        data=torch.ones((16, 32)),
        pixelscale=1.0,
        zeropoint=1.0,
    )

    # image reduction
    for scale in [2, 4, 8, 16]:
        reduced_image = new_image.reduce(scale)

        assert reduced_image.data[0][0] == scale**2, "reduced image should sum sub pixels"
        assert reduced_image.pixelscale == scale, "pixelscale should increase with reduced image"

    # image cropping
    crop_image = new_image.crop([1])
    assert crop_image.shape[1] == 14, "crop should cut 1 pixel from both sides here"
    crop_image = new_image.crop([3, 2])
    assert (
        crop_image.data.shape[0] == 26
    ), "crop should have cut 3 pixels from both sides of this axis"
    crop_image = new_image.crop([3, 2, 1, 0])
    assert (
        crop_image.data.shape[0] == 27
    ), "crop should have cut 3 pixels from left, 2 from right, 1 from top, and 0 from bottom"


def test_image_save_load():
    new_image = ap.Image(
        data=torch.ones((16, 32)),
        pixelscale=0.76,
        zeropoint=21.4,
        crtan=(8.0, 1.2),
        crpix=(2, 3),
        crval=(100.0, -32.1),
    )

    new_image.save("Test_AstroPhot.fits")

    loaded_image = ap.Image(filename="Test_AstroPhot.fits")

    assert torch.all(
        new_image.data == loaded_image.data
    ), "Loaded image should have same pixel values"
    assert torch.all(
        new_image.crtan.value == loaded_image.crtan.value
    ), "Loaded image should have same tangent plane origin"
    assert np.all(
        new_image.crpix == loaded_image.crpix
    ), "Loaded image should have same reference pixel"
    assert torch.all(
        new_image.crval.value == loaded_image.crval.value
    ), "Loaded image should have same reference world coordinates"
    assert torch.allclose(
        new_image.pixelscale, loaded_image.pixelscale
    ), "Loaded image should have same pixel scale"
    assert torch.allclose(
        new_image.CD.value, loaded_image.CD.value
    ), "Loaded image should have same pixel scale"
    assert new_image.zeropoint == loaded_image.zeropoint, "Loaded image should have same zeropoint"


def test_image_wcs_roundtrip():
    # Minimal input
    I = ap.Image(
        data=torch.zeros((21, 21)),
        zeropoint=22.5,
        crpix=(10, 10),
        crtan=(1.0, -10.0),
        crval=(160.0, 45.0),
        CD=0.05
        * np.array(
            [[np.cos(np.pi / 4), -np.sin(np.pi / 4)], [np.sin(np.pi / 4), np.cos(np.pi / 4)]]
        ),
    )

    assert torch.allclose(
        torch.stack(I.world_to_plane(*I.plane_to_world(*I.center))),
        I.center,
    ), "WCS world/plane roundtrip should return input value"
    assert torch.allclose(
        torch.stack(I.pixel_to_plane(*I.plane_to_pixel(*I.center))),
        I.center,
    ), "WCS pixel/plane roundtrip should return input value"
    assert torch.allclose(
        torch.stack(I.world_to_pixel(*I.pixel_to_world(*torch.zeros_like(I.center)))),
        torch.zeros_like(I.center),
        atol=1e-6,
    ), "WCS world/pixel roundtrip should return input value"


def test_target_image_variance():
    new_image = ap.TargetImage(
        data=torch.ones((16, 32)),
        variance=torch.ones((16, 32)),
        pixelscale=1.0,
        zeropoint=1.0,
    )

    assert new_image.has_variance, "target image should store variance"

    reduced_image = new_image.reduce(2)
    assert reduced_image.variance[0][0] == 4, "reduced image should sum sub pixels"

    new_image.variance = None
    assert not new_image.has_variance, "target image update to no variance"


def test_target_image_mask():
    new_image = ap.TargetImage(
        data=torch.ones((16, 32)),
        mask=torch.arange(16 * 32).reshape((16, 32)) % 4 == 0,
        pixelscale=1.0,
        zeropoint=1.0,
    )
    assert new_image.has_mask, "target image should store mask"

    reduced_image = new_image.reduce(2)
    assert reduced_image.mask[0][0] == 1, "reduced image should mask appropriately"
    assert reduced_image.mask[1][0] == 0, "reduced image should mask appropriately"

    new_image.mask = None
    assert not new_image.has_mask, "target image update to no mask"

    data = torch.ones((16, 32))
    data[1, 1] = torch.nan
    data[5, 5] = torch.nan

    new_image = ap.TargetImage(
        data=data,
        pixelscale=1.0,
        zeropoint=1.0,
    )
    assert new_image.has_mask, "target image with nans should create mask"
    assert new_image.mask[1][1].item() == True, "nan should be masked"
    assert new_image.mask[5][5].item() == True, "nan should be masked"


def test_target_image_psf():
    new_image = ap.TargetImage(
        data=torch.ones((15, 33)),
        psf=torch.ones((9, 9)),
        pixelscale=1.0,
        zeropoint=1.0,
    )
    assert new_image.has_psf, "target image should store variance"
    assert new_image.psf.psf_pad == 4, "psf border should be half psf size"

    reduced_image = new_image.reduce(3)
    assert reduced_image.psf.data[0][0] == 9, "reduced image should sum sub pixels in psf"

    new_image.psf = None
    assert not new_image.has_psf, "target image update to no variance"


def test_target_image_reduce():
    new_image = ap.TargetImage(
        data=torch.ones((30, 36)),
        psf=torch.ones((9, 9)),
        variance="auto",
        pixelscale=1.0,
        zeropoint=1.0,
    )
    smaller_image = new_image.reduce(3)
    assert smaller_image.data[0][0] == 9, "reduction should sum flux"
    assert tuple(smaller_image.data.shape) == (12, 10), "reduction should decrease image size"


def test_target_image_save_load():
    new_image = ap.TargetImage(
        data=torch.ones((16, 32)),
        variance=torch.ones((16, 32)),
        mask=torch.zeros((16, 32)),
        psf=torch.ones((9, 9)),
        CD=[[1.0, 0.0], [0.0, 1.5]],
        zeropoint=1.0,
    )

    new_image.save("Test_target_AstroPhot.fits")

    loaded_image = ap.TargetImage(filename="Test_target_AstroPhot.fits")

    assert torch.all(
        new_image.data == loaded_image.data
    ), "Loaded image should have same pixel values"
    assert torch.all(new_image.mask == loaded_image.mask), "Loaded image should have same mask"
    assert torch.all(
        new_image.variance == loaded_image.variance
    ), "Loaded image should have same variance"
    assert torch.all(
        new_image.psf.data == loaded_image.psf.data
    ), "Loaded image should have same psf"
    assert torch.allclose(
        new_image.CD.value, loaded_image.CD.value
    ), "Loaded image should have same pixel scale"


def test_target_image_auto_var():
    target = make_basic_sersic()
    target.variance = "auto"


def test_target_image_errors():
    new_image = ap.TargetImage(
        data=torch.ones((16, 32)),
        pixelscale=1.0,
        zeropoint=1.0,
    )

    # bad variance
    with pytest.raises(ap.errors.SpecificationConflict):
        new_image.variance = np.ones((5, 5))

    # bad mask
    with pytest.raises(ap.errors.SpecificationConflict):
        new_image.mask = np.zeros((5, 5))


def test_psf_image_copying():
    psf_image = ap.PSFImage(
        data=torch.ones((15, 15)),
    )

    assert psf_image.psf_pad == 7, "psf image should have correct psf_pad"
    psf_image.normalize()
    assert np.allclose(
        psf_image.data.detach().cpu().numpy(), 1 / 15**2
    ), "psf image should normalize to sum to 1"


def test_jacobian_add():
    new_image = ap.JacobianImage(
        parameters=["a", "b", "c"],
        data=torch.ones((16, 32, 3)),
    )
    other_image = ap.JacobianImage(
        parameters=["b", "d"],
        data=5 * torch.ones((4, 4, 2)),
    )

    new_image += other_image

    assert tuple(new_image.data.shape) == (
        32,
        16,
        3,
    ), "Jacobian addition should manage parameter identities"
    assert tuple(new_image.flatten("data").shape) == (
        512,
        3,
    ), "Jacobian should flatten to Npix*Nparams tensor"
    assert new_image.data[0, 0, 0].item() == 1, "Jacobian addition should not change original data"
    assert new_image.data[0, 0, 1].item() == 6, " Jacobian addition should add correctly"


def test_image_with_wcs():
    WCS = get_astropy_wcs()
    image = ap.TargetImage(
        data=np.ones((170, 180)),
        wcs=WCS,
    )
    assert image.shape[0] == WCS.pixel_shape[0], "Image should have correct shape from WCS"
    assert image.shape[1] == WCS.pixel_shape[1], "Image should have correct shape from WCS"
    assert np.allclose(
        image.CD.value * ap.utils.conversions.units.arcsec_to_deg, WCS.pixel_scale_matrix
    ), "Image should have correct CD from WCS"
    assert np.allclose(
        image.crpix, WCS.wcs.crpix[::-1] - 1
    ), "Image should have correct CRPIX from WCS"
    assert np.allclose(
        image.crval.value.detach().cpu().numpy(), WCS.wcs.crval
    ), "Image should have correct CRVAL from WCS"



import matplotlib
import matplotlib.pyplot as plt
import pytest


@pytest.fixture(autouse=True)
def no_block_show(monkeypatch):
    def close_show(*args, **kwargs):
        # plt.savefig("/dev/null")  # or do nothing
        plt.close("all")

    monkeypatch.setattr(plt, "show", close_show)

    # Also ensure we are in a non-GUI backend
    matplotlib.use("Agg")



import numpy as np
import torch

from .sky_model_object import SkyModel
from ..utils.decorators import ignore_numpy_warnings
from ..utils.interpolate import interp2d
from ..param import forward
from . import func
from ..utils.initialize import polar_decomposition

__all__ = ["BilinearSky"]


class BilinearSky(SkyModel):
    """Sky background model using a coarse bilinear grid for the sky flux.

    Parameters:
        I: sky brightness grid

    """

    _model_type = "bilinear"
    _parameter_specs = {
        "I": {"units": "flux/arcsec^2"},
        "PA": {"units": "radians", "shape": ()},
        "scale": {"units": "arcsec/grid-unit", "shape": ()},
    }
    sampling_mode = "midpoint"
    usable = True

    def __init__(self, *args, nodes=(3, 3), **kwargs):
        """Initialize the BilinearSky model with a grid of nodes."""
        super().__init__(*args, **kwargs)
        self.nodes = nodes

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if self.I.initialized:
            self.nodes = tuple(self.I.value.shape)

        if not self.PA.initialized:
            R, _ = polar_decomposition(self.target.CD.value.detach().cpu().numpy())
            self.PA.value = np.arccos(np.abs(R[0, 0]))
        if not self.scale.initialized:
            self.scale.value = (
                self.target.pixelscale.item() * self.target.data.shape[0] / self.nodes[0]
            )

        if self.I.initialized:
            return

        target_dat = self.target[self.window]
        dat = target_dat.data.detach().cpu().numpy().copy()
        if self.target.has_mask:
            mask = target_dat.mask.detach().cpu().numpy().copy()
            dat[mask] = np.nanmedian(dat)
        iS = dat.shape[0] // self.nodes[0]
        jS = dat.shape[1] // self.nodes[1]

        self.I.dynamic_value = (
            np.median(
                dat[: iS * self.nodes[0], : jS * self.nodes[1]].reshape(
                    iS, self.nodes[0], jS, self.nodes[1]
                ),
                axis=(0, 2),
            )
            / self.target.pixel_area.item()
        )

    @forward
    def transform_coordinates(self, x, y, I, PA, scale):
        x, y = super().transform_coordinates(x, y)
        i, j = func.rotate(-PA, x, y)
        pixel_center = (I.shape[0] - 1) / 2, (I.shape[1] - 1) / 2
        return i / scale + pixel_center[0], j / scale + pixel_center[1]

    @forward
    def brightness(self, x, y, I):
        x, y = self.transform_coordinates(x, y)
        return interp2d(I, y, x)



from typing import Optional, Sequence, Union

import torch
from caskade import forward

from .base import Model
from ..image import (
    Image,
    TargetImage,
    TargetImageList,
    ModelImage,
    ModelImageList,
    ImageList,
    Window,
    WindowList,
    JacobianImage,
    JacobianImageList,
)
from .. import config
from ..utils.decorators import ignore_numpy_warnings
from ..errors import InvalidTarget, InvalidWindow

__all__ = ["GroupModel"]


class GroupModel(Model):
    """Model object which represents a list of other models. For each
    general AstroPhot model method, this calls all the appropriate
    models from its list and combines their output into a single
    summed model. This class should be used when describing any
    system more complex than makes sense to represent with a single
    light distribution.

    Args:
        name (str): unique name for the full group model
        target (Target_Image): the target image that this group model is trying to fit to
        models (Optional[Sequence[AstroPhot_Model]]): list of AstroPhot_Model objects which will combine for the group model
        locked (bool): if the whole group of models should be locked

    """

    _model_type = "group"
    usable = True

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        models: Optional[Sequence[Model]] = None,
        **kwargs,
    ):
        super().__init__(name=name, **kwargs)
        for model in models:
            if not isinstance(model, Model):
                raise TypeError(f"Expected a Model instance in 'models', got {type(model)}")
        self.models = models
        self.update_window()

    def update_window(self):
        """Makes a new window object which encloses all the windows of the
        sub models in this group model object.

        """
        if isinstance(self.target, ImageList):  # WindowList if target is a TargetImageList
            new_window = list(target.window.copy() for target in self.target)
            n_windows = [0] * len(self.target.images)
            for model in self.models:
                if isinstance(model.target, ImageList):
                    for target, window in zip(model.target, model.window):
                        index = self.target.index(target)
                        if n_windows[index] == 0:
                            new_window[index] &= window
                        else:
                            new_window[index] |= window
                        n_windows[index] += 1
                elif isinstance(model.target, TargetImage):
                    index = self.target.index(model.target)
                    if n_windows[index] == 0:
                        new_window[index] &= model.window
                    else:
                        new_window[index] |= model.window
                    n_windows[index] += 1
                else:
                    raise NotImplementedError(
                        f"Group_Model cannot construct a window for itself using {type(model.target)} object. Must be a Target_Image"
                    )
            new_window = WindowList(new_window)
            for i, n in enumerate(n_windows):
                if n == 0:
                    config.logger.warning(
                        f"Model {self.name} has no sub models in target '{self.target.images[i].name}', this may cause issues with fitting."
                    )
        else:
            new_window = None
            for model in self.models:
                if new_window is None:
                    new_window = model.window.copy()
                else:
                    new_window |= model.window
        self.window = new_window

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        """
        Initialize each model in this group. Does this by iteratively initializing a model then subtracting it from a copy of the target.

        Args:
          target (Optional["Target_Image"]): A Target_Image instance to use as the source for initializing the model parameters on this image.
        """
        for model in self.models:
            config.logger.info(f"Initializing model {model.name}")
            model.initialize()

    def fit_mask(self) -> torch.Tensor:
        """Returns a mask for the target image which is the combination of all
        the fit masks of the sub models. This mask is used when the multiple
        models in the group model do not completely overlap with each other, thus
        there are some pixels which are not covered by any model and have no
        reason to be fit.

        """
        subtarget = self.target[self.window]
        if isinstance(subtarget, ImageList):
            mask = tuple(torch.ones_like(submask) for submask in subtarget.mask)
            for model in self.models:
                model_subtarget = model.target[model.window]
                model_fit_mask = model.fit_mask()
                if isinstance(model_subtarget, ImageList):
                    for target, submask in zip(model_subtarget, model_fit_mask):
                        index = subtarget.index(target)
                        group_indices = subtarget.images[index].get_indices(target.window)
                        model_indices = target.get_indices(subtarget.images[index].window)
                        mask[index][group_indices] &= submask[model_indices]
                else:
                    index = subtarget.index(model_subtarget)
                    group_indices = subtarget.images[index].get_indices(model_subtarget.window)
                    model_indices = model_subtarget.get_indices(subtarget.images[index].window)
                    mask[index][group_indices] &= model_fit_mask[model_indices]
        else:
            mask = torch.ones_like(subtarget.mask)
            for model in self.models:
                model_subtarget = model.target[model.window]
                group_indices = subtarget.get_indices(model.window)
                model_indices = model_subtarget.get_indices(subtarget.window)
                mask[group_indices] &= model.fit_mask()[model_indices]
        return mask

    def match_window(self, image, window, model):
        if isinstance(image, ImageList) and isinstance(model.target, ImageList):
            indices = image.match_indices(model.target)
            if len(indices) == 0:
                raise IndexError
            use_window = WindowList(windows=list(image.images[i].window for i in indices))
        elif isinstance(image, ImageList) and isinstance(model.target, Image):
            try:
                image.index(model.target)
            except ValueError:
                raise IndexError
            use_window = model.window
        elif isinstance(image, Image) and isinstance(model.target, ImageList):
            try:
                i = model.target.index(image)
            except ValueError:
                raise IndexError
            use_window = model.window[i]
        elif isinstance(image, Image) and isinstance(model.target, Image):
            if image.identity != model.target.identity:
                raise IndexError
            use_window = window
        else:
            raise NotImplementedError(
                f"Group_Model cannot sample with {type(image)} and {type(model.target)}"
            )
        return use_window

    def _ensure_vmap_compatible(self, image, other):
        if isinstance(image, ImageList):
            for img in image.images:
                self._ensure_vmap_compatible(img, other)
            return
        if isinstance(other, ImageList):
            for img in other.images:
                self._ensure_vmap_compatible(image, img)
            return
        if image.identity == other.identity:
            image += torch.zeros_like(other.data[0, 0])

    @forward
    def sample(
        self,
        window: Optional[Window] = None,
    ) -> Union[ModelImage, ModelImageList]:
        """Sample the group model on an image. Produces the flux values for
        each pixel associated with the models in this group. Each
        model is called individually and the results are added
        together in one larger image.

        Args:
          image (Optional["Model_Image"]): Image to sample on, overrides the windows for each sub model, they will all be evaluated over this entire image. If left as none then each sub model will be evaluated in its window.

        """
        if window is None:
            image = self.target[self.window].model_image()
        else:
            image = self.target[window].model_image()

        for model in self.models:
            if window is None:
                use_window = model.window
            else:
                try:
                    use_window = self.match_window(image, window, model)
                except IndexError:
                    # If the model target is not in the image, skip it
                    continue
            model_image = model(window=model.window & use_window)
            self._ensure_vmap_compatible(image, model_image)
            image += model_image

        return image

    @torch.no_grad()
    def jacobian(
        self,
        pass_jacobian: Optional[Union[JacobianImage, JacobianImageList]] = None,
        window: Optional[Union[Window, WindowList]] = None,
        params=None,
    ) -> JacobianImage:
        """Compute the jacobian for this model. Done by first constructing a
        full jacobian (Npixels * Nparameters) of zeros then call the
        jacobian method of each sub model and add it in to the total.

        Args:
          pass_jacobian (Optional["Jacobian_Image"]): A Jacobian image pre-constructed to be passed along instead of constructing new Jacobians

        """
        if window is None:
            window = self.window

        if params is not None:
            self.fill_dynamic_values(params)

        if pass_jacobian is None:
            jac_img = self.target[window].jacobian_image(
                parameters=self.build_params_array_identities()
            )
        else:
            jac_img = pass_jacobian

        for model in reversed(self.models):
            try:
                use_window = self.match_window(jac_img, window, model)
            except IndexError:
                # If the model target is not in the image, skip it
                continue
            jac_img = model.jacobian(
                pass_jacobian=jac_img,
                window=use_window & model.window,
            )

        return jac_img

    def __iter__(self):
        return (mod for mod in self.models)

    @property
    def target(self) -> Optional[Union[TargetImage, TargetImageList]]:
        try:
            return self._target
        except AttributeError:
            return None

    @target.setter
    def target(self, tar: Optional[Union[TargetImage, TargetImageList]]):
        if not (tar is None or isinstance(tar, (TargetImage, TargetImageList))):
            raise InvalidTarget("Group_Model target must be a Target_Image instance.")
        try:
            del self._target  # Remove old target if it exists
        except AttributeError:
            pass

        self._target = tar

    @property
    def window(self) -> Optional[Union[Window, WindowList]]:
        """The window defines a region on the sky in which this model will be
        optimized and typically evaluated. Two models with
        non-overlapping windows are in effect independent of each
        other. If there is another model with a window that spans both
        of them, then they are tenuously connected.

        If not provided, the model will assume a window equal to the
        target it is fitting. Note that in this case the window is not
        explicitly set to the target window, so if the model is moved
        to another target then the fitting window will also change.

        """
        if self._window is None:
            if self.target is None:
                raise ValueError(
                    "This model has no target or window, these must be provided by the user"
                )
            return self.target.window
        return self._window

    @window.setter
    def window(self, window):
        if window is None:
            self._window = None
        elif isinstance(window, (Window, WindowList)):
            self._window = window
        elif len(window) in [2, 4]:
            self._window = Window(window, image=self.target)
        else:
            raise InvalidWindow(f"Unrecognized window format: {str(window)}")



import torch


def euler_rotation_matrix(alpha, beta, gamma):
    """Compute the rotation matrix from Euler angles.

    See the Z_alpha X_beta Z_gamma convention for the order of rotations here:
    https://en.wikipedia.org/wiki/Euler_angles
    """
    ca = torch.cos(alpha)
    sa = torch.sin(alpha)
    cb = torch.cos(beta)
    sb = torch.sin(beta)
    cg = torch.cos(gamma)
    sg = torch.sin(gamma)
    R = torch.stack(
        (
            torch.stack((ca * cg - cb * sa * sg, -ca * sg - cb * cg * sa, sb * sa)),
            torch.stack((cg * sa + ca * cb * sg, ca * cb * cg - sa * sg, -ca * sb)),
            torch.stack((sb * cg, sb * cg, cb)),
        ),
        dim=-1,
    )
    return R



# Apply an optimizer toa  downsampled version of an image
from typing import Dict, Any

import numpy as np

from .base import BaseOptimizer
from ..models import Model
from .lm import LM
from .. import config

__all__ = ["MiniFit"]


class MiniFit(BaseOptimizer):
    def __init__(
        self,
        model: Model,
        downsample_factor: int = 2,
        max_pixels: int = 10000,
        method: BaseOptimizer = LM,
        initial_state: np.ndarray = None,
        method_kwargs: Dict[str, Any] = {},
        **kwargs: Dict[str, Any],
    ) -> None:
        super().__init__(model, initial_state, **kwargs)

        self.method = method
        self.method_kwargs = method_kwargs
        if "verbose" not in self.method_kwargs:
            self.method_kwargs["verbose"] = self.verbose

        self.downsample_factor = downsample_factor
        self.max_pixels = max_pixels

    def fit(self) -> BaseOptimizer:
        initial_target = self.model.target
        target_area = self.model.target[self.model.window]
        while True:
            small_target = target_area.reduce(self.downsample_factor)
            if np.prod(small_target.shape) < self.max_pixels:
                break
            self.downsample_factor += 1

        if self.verbose > 0:
            config.logger.info(f"Downsampling target by {self.downsample_factor}x")

        self.small_target = small_target
        self.model.target = small_target
        res = self.method(self.model, **self.method_kwargs).fit()
        self.model.target = initial_target

        self.message = res.message

        return self



from functools import partial
from typing import Literal

import numpy as np
import torch
from scipy.stats import binned_statistic, iqr

from .. import config
from ..models import Model

# from ..models import Warp_Galaxy
from ..utils.conversions.units import flux_to_sb
from .visuals import *

__all__ = [
    "radial_light_profile",
    "radial_median_profile",
    "ray_light_profile",
    "wedge_light_profile",
    "warp_phase_profile",
]


def radial_light_profile(
    fig,
    ax,
    model: Model,
    rad_unit="arcsec",
    extend_profile=1.0,
    R0=0.0,
    resolution=1000,
    plot_kwargs={},
):
    xx = torch.linspace(
        R0,
        max(model.window.shape)
        * model.target.pixelscale.detach().cpu().numpy()
        * extend_profile
        / 2,
        int(resolution),
        dtype=config.DTYPE,
        device=config.DEVICE,
    )
    flux = model.radial_model(xx, params=()).detach().cpu().numpy()
    if model.target.zeropoint is not None:
        yy = flux_to_sb(flux, 1.0, model.target.zeropoint.item())
    else:
        yy = np.log10(flux)

    kwargs = {
        "linewidth": 2,
        "color": main_pallet["primary1"],
        "label": f"{model.name} profile",
    }
    kwargs.update(plot_kwargs)
    with torch.no_grad():
        ax.plot(
            xx.detach().cpu().numpy(),
            yy,
            **kwargs,
        )

    if model.target.zeropoint is not None:
        ax.set_ylabel("Surface Brightness [mag/arcsec$^2$]")
        if not ax.yaxis_inverted():
            ax.invert_yaxis()
    else:
        ax.set_ylabel("log$_{10}$(flux/arcsec$^2$)")
    ax.set_xlabel(f"Radius [{rad_unit}]")
    ax.set_xlim([R0, None])
    return fig, ax


def radial_median_profile(
    fig,
    ax,
    model: Model,
    count_limit: int = 10,
    return_profile: bool = False,
    rad_unit: str = "arcsec",
    plot_kwargs: dict = {},
):
    """Plot an SB profile by taking flux median at each radius.

    Using the coordinate transforms defined by the model object,
    assigns a radius to each pixel then bins the pixel-radii and
    computes the median in each bin. This gives a simple
    representation of the image data if one were to simply average the
    pixels along isophotes.

    Args:
      fig: matplotlib figure object
      ax: matplotlib axis object
      model (AstroPhot_Model): Model object from which to determine the radial binning. Also provides the target image to extract the data
      count_limit (int): The limit of pixels in a bin, below which uncertainties are not computed. Default: 10
      return_profile (bool): Instead of just returning the fig and ax object, will return the extracted profile formatted as: Rbins (the radial bin edges), medians (the median in each bin), scatter (the 16-84 quartile range / 2), count (the number of pixels in each bin). Default: False
      rad_unit (str): The name of the radius units to plot. If you select "pixel" then the plot will work in pixel units (physical radii divided by pixelscale) if you choose any other string then it will remain in the physical units of the image and the axis label will be whatever you set the value to. Default: "arcsec". Options: "arcsec", "pixel"
      bin_scale (float): The geometric scaling factor for the binning, each bin will be this much larger than the previous. Default: 0.1
      min_bin_width (float): The minimum width of a bin in pixel units, default is 2 so that each bin will have some data to compute the median with. Default: 2
      doassert (bool): If any requirements are imposed on which kind of profile can be plotted, this activates them. Default: True

    """

    Rlast_pix = max(model.window.shape) / 2
    Rlast_phys = Rlast_pix * model.target.pixelscale.item()

    Rbins = [0.0]
    while Rbins[-1] < Rlast_phys:
        Rbins.append(Rbins[-1] + max(2 * model.target.pixelscale.item(), Rbins[-1] * 0.1))
    Rbins = np.array(Rbins)

    with torch.no_grad():
        image = model.target[model.window]
        x, y = image.coordinate_center_meshgrid()
        x, y = model.transform_coordinates(x, y, params=())
        R = (x**2 + y**2).sqrt()
        R = R.detach().cpu().numpy()

    dat = image.data.detach().cpu().numpy()
    count, bins, binnum = binned_statistic(
        R.ravel(),
        dat.ravel(),
        statistic="count",
        bins=Rbins,
    )

    stat, bins, binnum = binned_statistic(
        R.ravel(),
        dat.ravel(),
        statistic="median",
        bins=Rbins,
    )
    stat[count == 0] = np.nan

    scat, bins, binnum = binned_statistic(
        R.ravel(),
        dat.ravel(),
        statistic=partial(iqr, rng=(16, 84)),
        bins=Rbins,
    )
    scat[count > count_limit] /= 2 * np.sqrt(count[count > count_limit])
    scat[count <= count_limit] = 0

    if model.target.zeropoint is not None:
        stat = flux_to_sb(stat, model.target.pixel_area.item(), model.target.zeropoint.item())
        ax.set_ylabel("Surface Brightness [mag/arcsec$^2$]")
        if not ax.yaxis_inverted():
            ax.invert_yaxis()
    else:
        stat = np.log10(stat)
        ax.set_ylabel("log$_{10}$(flux/arcsec^2)")

    kwargs = {
        "linewidth": 0,
        "elinewidth": 1,
        "color": main_pallet["primary2"],
        "label": "data profile",
        **plot_kwargs,
    }
    ax.errorbar(
        (Rbins[:-1] + Rbins[1:]) / 2,
        stat,
        yerr=scat,
        fmt=".",
        **kwargs,
    )
    ax.set_xlabel(f"Radius [{rad_unit}]")

    if return_profile:
        return Rbins, stat, scat, count
    return fig, ax


def ray_light_profile(
    fig,
    ax,
    model: Model,
    rad_unit="arcsec",
    extend_profile=1.0,
    resolution=1000,
):
    xx = torch.linspace(
        0,
        max(model.window.shape) * model.target.pixelscale * extend_profile / 2,
        int(resolution),
        dtype=config.DTYPE,
        device=config.DEVICE,
    )
    for r in range(model.segments):
        if model.segments <= 3:
            col = main_pallet[f"primary{r+1}"]
        else:
            col = cmap_grad(r / model.segments)
        with torch.no_grad():
            ax.plot(
                xx.detach().cpu().numpy(),
                np.log10(model.iradial_model(r, xx, params=()).detach().cpu().numpy()),
                linewidth=2,
                color=col,
                label=f"{model.name} profile {r}",
            )
    ax.set_ylabel("log$_{10}$(flux)")
    ax.set_xlabel(f"Radius [{rad_unit}]")

    return fig, ax


def wedge_light_profile(
    fig,
    ax,
    model: Model,
    rad_unit="arcsec",
    extend_profile=1.0,
    resolution=1000,
):
    xx = torch.linspace(
        0,
        max(model.window.shape) * model.target.pixelscale * extend_profile / 2,
        int(resolution),
        dtype=config.DTYPE,
        device=config.DEVICE,
    )
    for r in range(model.segments):
        if model.segments <= 3:
            col = main_pallet[f"primary{r+1}"]
        else:
            col = cmap_grad(r / model.segments)
        with torch.no_grad():
            ax.plot(
                xx.detach().cpu().numpy(),
                np.log10(model.iradial_model(r, xx, params=()).detach().cpu().numpy()),
                linewidth=2,
                color=col,
                label=f"{model.name} profile {r}",
            )
    ax.set_ylabel("log$_{10}$(flux)")
    ax.set_xlabel(f"Radius [{rad_unit}]")

    return fig, ax


def warp_phase_profile(fig, ax, model: Model, rad_unit="arcsec"):

    ax.plot(
        model.q_R.prof.detach().cpu().numpy(),
        model.q_R.npvalue,
        linewidth=2,
        color=main_pallet["primary1"],
        label=f"{model.name} axis ratio",
    )
    ax.plot(
        model.PA_R.prof.detach().cpu().numpy(),
        model.PA_R.npvalue / np.pi,
        linewidth=2,
        color=main_pallet["primary2"],
        label=f"{model.name} position angle/$\\pi$",
    )
    ax.set_ylim([0, 1])
    ax.set_ylabel("q [b/a], PA [rad/$\\pi$]")
    ax.set_xlabel(f"Radius [{rad_unit}]")

    return fig, ax



import torch

from .. import config


def chi_squared(target, model, mask=None, variance=None):
    if mask is None:
        if variance is None:
            return torch.sum((target - model) ** 2)
        else:
            return torch.sum(((target - model) ** 2) / variance)
    else:
        mask = torch.logical_not(mask)
        if variance is None:
            return torch.sum((target[mask] - model[mask]) ** 2)
        else:
            return torch.sum(((target[mask] - model[mask]) ** 2) / variance[mask])


def reduced_chi_squared(target, model, params, mask=None, variance=None):
    if mask is None:
        ndf = (
            torch.prod(torch.tensor(target.shape, dtype=config.DTYPE, device=config.DEVICE))
            - params
        )
    else:
        ndf = torch.sum(torch.logical_not(mask)) - params
    return chi_squared(target, model, mask, variance) / ndf



import torch
import numpy as np

from ...param import forward
from ...utils.decorators import ignore_numpy_warnings
from ...utils.parametric_profiles import king_np
from .._shared_methods import parametric_initialize, parametric_segment_initialize
from .. import func


def x0_func(model_params, R, F):
    return R[2], R[5], 2, 10 ** F[0]


class KingMixin:
    """Empirical King radial light profile (Elson 1999).

    Often used for star clusters. By default the profile has `alpha = 2` but we
    allow the parameter to vary freely for fitting. The functional form of the
    Empirical King profile is defined as:

    $$I(R) = I_0\\left[\\frac{1}{(1 + (R/R_c)^2)^{1/\\alpha}} - \\frac{1}{(1 + (R_t/R_c)^2)^{1/\\alpha}}\\right]^{\\alpha}\\left[1 - \\frac{1}{(1 + (R_t/R_c)^2)^{1/\\alpha}}\\right]^{-\\alpha}$$

    where `R_c` is the core radius, `R_t` is the truncation radius, and `I_0` is
    the intensity at the center of the profile. `alpha` is the concentration
    index which controls the shape of the profile.

    Parameters:
        Rc: core radius
        Rt: truncation radius
        alpha: concentration index which controls the shape of the brightness profile
        I0: intensity at the center of the profile
    """

    _model_type = "king"
    _parameter_specs = {
        "Rc": {"units": "arcsec", "valid": (0.0, None), "shape": ()},
        "Rt": {"units": "arcsec", "valid": (0.0, None), "shape": ()},
        "alpha": {"units": "unitless", "valid": (0, 10), "shape": (), "value": 2.0},
        "I0": {"units": "flux/arcsec^2", "valid": (0, None), "shape": ()},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if not self.alpha.initialized:
            self.alpha.dynamic_value = 2.0

        parametric_initialize(
            self,
            self.target[self.window],
            lambda r, *x: king_np(r, x[0], x[1], 2.0, x[2]),
            ("Rc", "Rt", "I0"),
            x0_func,
        )

    @forward
    def radial_model(self, R, Rc, Rt, alpha, I0):
        return func.king(R, Rc, Rt, alpha, I0)


class iKingMixin:
    """Empirical King radial light profile (Elson 1999).

    Often used for star clusters. By default the profile has `alpha = 2` but we
    allow the parameter to vary freely for fitting. The functional form of the
    Empirical King profile is defined as:

    $$I(R) = I_0\\left[\\frac{1}{(1 + (R/R_c)^2)^{1/\\alpha}} - \\frac{1}{(1 + (R_t/R_c)^2)^{1/\\alpha}}\\right]^{\\alpha}\\left[1 - \\frac{1}{(1 + (R_t/R_c)^2)^{1/\\alpha}}\\right]^{-\\alpha}$$

    where `R_c` is the core radius, `R_t` is the truncation radius, and `I_0` is
    the intensity at the center of the profile. `alpha` is the concentration
    index which controls the shape of the profile.

    `Rc`, `Rt`, `alpha`, and `I0` are batched by their first dimension, allowing
    for multiple King profiles to be defined at once.

    Parameters:
        Rc: core radius
        Rt: truncation radius
        alpha: concentration index which controls the shape of the brightness profile
        I0: intensity at the center of the profile
    """

    _model_type = "king"
    _parameter_specs = {
        "Rc": {"units": "arcsec", "valid": (0.0, None)},
        "Rt": {"units": "arcsec", "valid": (0.0, None)},
        "alpha": {"units": "unitless", "valid": (0, 10)},
        "I0": {"units": "flux/arcsec^2", "valid": (0, None)},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if not self.alpha.initialized:
            self.alpha.value = 2.0 * np.ones(self.segments)
        parametric_segment_initialize(
            model=self,
            target=self.target[self.window],
            prof_func=lambda r, *x: king_np(r, x[0], x[1], 2.0, x[2]),
            params=("Rc", "Rt", "I0"),
            x0_func=x0_func,
            segments=self.segments,
        )

    @forward
    def iradial_model(self, i, R, Rc, Rt, alpha, I0):
        return func.king(R, Rc[i], Rt[i], alpha[i], I0[i])



from typing import Optional, Literal

import numpy as np
from torch.autograd.functional import jacobian
from torch.func import jacfwd, hessian
import torch
from torch import Tensor

from ...param import forward
from ... import config
from ...image import Image, Window, JacobianImage
from .. import func
from ...errors import SpecificationConflict


class SampleMixin:
    """
    options:
        sampling_mode: The method used to sample the model in image pixels. Options are:
            - auto: Automatically choose the sampling method based on the image size.
            - midpoint: Use midpoint sampling, evaluate the brightness at the center of each pixel.
            - simpsons: Use Simpson's rule for sampling integrating each pixel.
            - quad:x: Use quadrature sampling with order x, where x is a positive integer to integrate each pixel.
        jacobian_maxparams: The maximum number of parameters before the Jacobian will be broken into
            smaller chunks. This is helpful for limiting the memory requirements to build a model.
        jacobian_maxpixels: The maximum number of pixels before the Jacobian will be broken into
            smaller chunks. This is helpful for limiting the memory requirements to build a model.
        integrate_mode: The method used to select pixels to integrate further where the model varies significantly. Options are:
            - none: No extra integration is performed (beyond the sampling_mode).
            - bright: Select the brightest pixels for further integration.
            - threshold: Select pixels which show signs of significant  higher order derivatives.
        integrate_tolerance: The tolerance for selecting a pixel in the integration method. This is the total flux fraction
            that is integrated over the image.
        integrate_fraction: The fraction of the pixels to super sample during integration.
        integrate_max_depth: The maximum depth of the integration method.
        integrate_gridding: The gridding used for the integration method to super-sample a pixel at each iteration.
        integrate_quad_order: The order of the quadrature used for the integration method on the super sampled pixels.
    """

    # Method for initial sampling of model
    sampling_mode = "auto"  # auto (choose based on image size), midpoint, simpsons, quad:x (where x is a positive integer)

    # Maximum size of parameter list before jacobian will be broken into smaller chunks, this is helpful for limiting the memory requirements to build a model, lower jacobian_chunksize is slower but uses less memory
    jacobian_maxparams = 10
    jacobian_maxpixels = 1000**2
    integrate_mode = "bright"  # none, bright, threshold
    integrate_tolerance = 1e-4  # total flux fraction
    integrate_fraction = 0.05  # fraction of the pixels to super sample
    integrate_max_depth = 2
    integrate_gridding = 5
    integrate_quad_order = 3

    _options = (
        "sampling_mode",
        "jacobian_maxparams",
        "jacobian_maxpixels",
        "integrate_mode",
        "integrate_tolerance",
        "integrate_fraction",
        "integrate_max_depth",
        "integrate_gridding",
        "integrate_quad_order",
    )

    @forward
    def _bright_integrate(self, sample, image):
        i, j = image.pixel_center_meshgrid()
        N = max(1, int(np.prod(image.data.shape) * self.integrate_fraction))
        sample_flat = sample.flatten(-2)
        select = torch.topk(sample_flat, N, dim=-1).indices
        sample_flat[select] = func.recursive_bright_integrate(
            i.flatten(-2)[select],
            j.flatten(-2)[select],
            lambda i, j: self.brightness(*image.pixel_to_plane(i, j)),
            bright_frac=self.integrate_fraction,
            quad_order=self.integrate_quad_order,
            gridding=self.integrate_gridding,
            max_depth=self.integrate_max_depth,
        )
        return sample_flat.reshape(sample.shape)

    @forward
    def _threshold_integrate(self, sample, image: Image):
        i, j = image.pixel_center_meshgrid()
        kernel = func.curvature_kernel(config.DTYPE, config.DEVICE)
        curvature = (
            torch.nn.functional.pad(
                torch.nn.functional.conv2d(
                    sample.view(1, 1, *sample.shape),
                    kernel.view(1, 1, *kernel.shape),
                    padding="valid",
                ),
                (1, 1, 1, 1),
                mode="replicate",
            )
            .squeeze(0)
            .squeeze(0)
            .abs()
        )
        total_est = torch.sum(sample)
        threshold = total_est * self.integrate_tolerance
        select = curvature > threshold

        sample[select] = func.recursive_quad_integrate(
            i[select],
            j[select],
            lambda i, j: self.brightness(*image.pixel_to_plane(i, j)),
            threshold=threshold,
            quad_order=self.integrate_quad_order,
            gridding=self.integrate_gridding,
            max_depth=self.integrate_max_depth,
        )
        return sample

    @forward
    def sample_image(self, image: Image):
        if self.sampling_mode == "auto":
            N = np.prod(image.data.shape)
            if N <= 100:
                sampling_mode = "quad:5"
            elif N <= 10000:
                sampling_mode = "simpsons"
            else:
                sampling_mode = "midpoint"
        else:
            sampling_mode = self.sampling_mode
        if sampling_mode == "midpoint":
            x, y = image.coordinate_center_meshgrid()
            res = self.brightness(x, y)
            sample = func.pixel_center_integrator(res)
        elif sampling_mode == "simpsons":
            x, y = image.coordinate_simpsons_meshgrid()
            res = self.brightness(x, y)
            sample = func.pixel_simpsons_integrator(res)
        elif sampling_mode.startswith("quad:"):
            order = int(self.sampling_mode.split(":")[1])
            i, j, w = image.pixel_quad_meshgrid(order=order)
            x, y = image.pixel_to_plane(i, j)
            res = self.brightness(x, y)
            sample = func.pixel_quad_integrator(res, w)
        else:
            raise SpecificationConflict(
                f"Unknown sampling mode {self.sampling_mode} for model {self.name}"
            )
        if self.integrate_mode == "threshold":
            sample = self._threshold_integrate(sample, image)
        elif self.integrate_mode == "bright":
            sample = self._bright_integrate(sample, image)
        elif self.integrate_mode != "none":
            raise SpecificationConflict(
                f"Unknown integrate mode {self.integrate_mode} for model {self.name}"
            )
        return sample

    def _jacobian(self, window: Window, params_pre: Tensor, params: Tensor, params_post: Tensor):
        # return jacfwd( # this should be more efficient, but the trace overhead is too high
        #     lambda x: self.sample(
        #         window=window, params=torch.cat((params_pre, x, params_post), dim=-1)
        #     ).data
        # )(params)
        return jacobian(
            lambda x: self.sample(
                window=window, params=torch.cat((params_pre, x, params_post), dim=-1)
            ).data,
            params,
            strategy="forward-mode",
            vectorize=True,
            create_graph=False,
        )

    def jacobian(
        self,
        window: Optional[Window] = None,
        pass_jacobian: Optional[JacobianImage] = None,
        params: Optional[Tensor] = None,
    ):
        if window is None:
            window = self.window

        if pass_jacobian is None:
            jac_img = self.target[window].jacobian_image(
                parameters=self.build_params_array_identities()
            )
        else:
            jac_img = pass_jacobian

        # No dynamic params
        if params is None:
            params = self.build_params_array()
        if params.shape[-1] == 0:
            return jac_img

        # handle large images
        n_pixels = np.prod(window.shape)
        if n_pixels > self.jacobian_maxpixels:
            for chunk in window.chunk(self.jacobian_maxpixels):
                jac_img = self.jacobian(window=chunk, pass_jacobian=jac_img, params=params)
            return jac_img

        identities = self.build_params_array_identities()
        if len(jac_img.match_parameters(identities)[0]) == 0:
            return jac_img

        target = self.target[window]
        if len(params) > self.jacobian_maxparams:  # handle large number of parameters
            chunksize = len(params) // self.jacobian_maxparams + 1
            for i in range(0, len(params), chunksize):
                params_pre = params[:i]
                params_chunk = params[i : i + chunksize]
                params_post = params[i + chunksize :]
                jac_chunk = self._jacobian(window, params_pre, params_chunk, params_post)
                jac_img += target.jacobian_image(
                    parameters=identities[i : i + chunksize],
                    data=jac_chunk,
                )
        else:
            jac = self._jacobian(window, params[:0], params, params[0:0])
            jac_img += target.jacobian_image(parameters=identities, data=jac)

        return jac_img

    def gradient(
        self,
        window: Optional[Window] = None,
        params: Optional[Tensor] = None,
        likelihood: Literal["gaussian", "poisson"] = "gaussian",
    ):
        """Compute the gradient of the model with respect to its parameters."""
        if window is None:
            window = self.window

        jacobian_image = self.jacobian(window=window, params=params)

        data = self.target[window].data
        model = self.sample(window=window).data
        if likelihood == "gaussian":
            weight = self.target[window].weight
            gradient = torch.sum(
                jacobian_image.data * ((data - model) * weight).unsqueeze(-1), dim=(0, 1)
            )
        elif likelihood == "poisson":
            gradient = torch.sum(
                jacobian_image.data * (1 - data / model).unsqueeze(-1),
                dim=(0, 1),
            )

        return gradient



# Levenberg-Marquardt algorithm
from typing import Sequence

import torch
import numpy as np

from .base import BaseOptimizer
from .. import config
from . import func
from ..errors import OptimizeStopFail, OptimizeStopSuccess
from ..param import ValidContext

__all__ = ("LM",)


class LM(BaseOptimizer):
    """The LM class is an implementation of the Levenberg-Marquardt
    optimization algorithm. This method is used to solve non-linear
    least squares problems and is known for its robustness and
    efficiency.

    The Levenberg-Marquardt (LM) algorithm is an iterative method used
    for solving non-linear least squares problems. It can be seen as a
    combination of the Gauss-Newton method and the gradient descent
    method: it works like the Gauss-Newton method when the parameters
    are approximately near a minimum, and like the gradient descent method
    when the parameters are far from their optimal values.

    The cost function that the LM algorithm tries to minimize is of
    the form:

    .. math::
        f(\\boldsymbol{\\beta}) = \\frac{1}{2}\\sum_{i=1}^{N} r_i(\\boldsymbol{\\beta})^2

    where :math:`\\boldsymbol{\\beta}` is the vector of parameters,
    :math:`r_i` are the residuals, and :math:`N` is the number of
    observations.

    The LM algorithm iteratively performs the following update to the parameters:

    .. math::
        \\boldsymbol{\\beta}_{n+1} = \\boldsymbol{\\beta}_{n} - (J^T J + \\lambda diag(J^T J))^{-1} J^T \\boldsymbol{r}

    where:
        - :math:`J` is the Jacobian matrix whose elements are :math:`J_{ij} = \\frac{\\partial r_i}{\\partial \\beta_j}`,
        - :math:`\\boldsymbol{r}` is the vector of residuals :math:`r_i(\\boldsymbol{\\beta})`,
        - :math:`\\lambda` is a damping factor which is adjusted at each iteration.

    When :math:`\\lambda = 0` this can be seen as the Gauss-Newton
    method. In the limit that :math:`\\lambda` is large, the
    :math:`J^T J` matrix (an approximation of the Hessian) becomes
    subdominant and the update essentially points along :math:`J^T
    \\boldsymbol{r}` which is the gradient. In this scenario the
    gradient descent direction is also modified by the :math:`\\lambda
    diag(J^T J)` scaling which in some sense makes each gradient
    unitless and further improves the step. Note as well that as
    :math:`\\lambda` gets larger the step taken will be smaller, which
    helps to ensure convergence when the initial guess of the
    parameters are far from the optimal solution.

    Note that the residuals :math:`r_i` are typically also scaled by
    the variance of the pixels, but this does not change the equations
    above. For a detailed explanation of the LM method see the article
    by Henri Gavin on which much of the AstroPhot LM implementation is
    based::

        @article{Gavin2019,
            title={The Levenberg-Marquardt algorithm for nonlinear least squares curve-fitting problems},
            author={Gavin, Henri P},
            journal={Department of Civil and Environmental Engineering, Duke University},
            volume={19},
            year={2019}
        }

    as well as the paper on LM geodesic acceleration by Mark
    Transtrum::

        @article{Tanstrum2012,
           author = {{Transtrum}, Mark K. and {Sethna}, James P.},
            title = "{Improvements to the Levenberg-Marquardt algorithm for nonlinear least-squares minimization}",
             year = 2012,
              doi = {10.48550/arXiv.1201.5885},
           adsurl = {https://ui.adsabs.harvard.edu/abs/2012arXiv1201.5885T},
        }

    The damping factor :math:`\\lambda` is adjusted at each iteration:
    it is effectively increased when we are far from the solution, and
    decreased when we are close to it. In practice, the algorithm
    attempts to pick the smallest :math:`\\lambda` that is can while
    making sure that the :math:`\\chi^2` decreases at each step.

    The main advantage of the LM algorithm is its adaptability. When
    the current estimate is far from the optimum, the algorithm
    behaves like a gradient descent to ensure global
    convergence. However, when it is close to the optimum, it behaves
    like Gauss-Newton, which provides fast local convergence.

    In practice, the algorithm is typically implemented with various
    enhancements to improve its performance. For example, the Jacobian
    may be approximated with finite differences, geodesic acceleration
    can be used to speed up convergence, and more sophisticated
    strategies can be used to adjust the damping factor :math:`\\lambda`.

    The exact performance of the LM algorithm will depend on the
    nature of the problem, including the complexity of the function
    f(x), the quality of the initial estimate x0, and the distribution
    of the data.

    The LM class implemented for AstroPhot takes a model, initial
    state, and various other optional parameters as inputs and seeks
    to find the parameters that minimize the cost function.

    """

    def __init__(
        self,
        model,
        initial_state: Sequence = None,
        max_iter: int = 100,
        relative_tolerance: float = 1e-5,
        Lup=11.0,
        Ldn=9.0,
        L0=1.0,
        max_step_iter: int = 10,
        ndf=None,
        **kwargs,
    ):

        super().__init__(
            model,
            initial_state,
            max_iter=max_iter,
            relative_tolerance=relative_tolerance,
            **kwargs,
        )
        # Maximum number of iterations of the algorithm
        self.max_iter = max_iter
        # Maximum number of steps while searching for chi^2 improvement on a single jacobian evaluation
        self.max_step_iter = max_step_iter
        self.Lup = Lup
        self.Ldn = Ldn
        self.L = L0

        # mask
        fit_mask = self.model.fit_mask()
        if isinstance(fit_mask, tuple):
            fit_mask = torch.cat(tuple(FM.flatten() for FM in fit_mask))
        else:
            fit_mask = fit_mask.flatten()
        if torch.sum(fit_mask).item() == 0:
            fit_mask = None

        if model.target.has_mask:
            mask = self.model.target[self.fit_window].flatten("mask")
            if fit_mask is not None:
                mask = mask | fit_mask
            self.mask = ~mask
        elif fit_mask is not None:
            self.mask = ~fit_mask
        else:
            self.mask = torch.ones_like(
                self.model.target[self.fit_window].flatten("data"), dtype=torch.bool
            )
        if self.mask is not None and torch.sum(self.mask).item() == 0:
            raise OptimizeStopSuccess("No data to fit. All pixels are masked")

        # Initialize optimizer attributes
        self.Y = self.model.target[self.fit_window].flatten("data")[self.mask]

        # 1 / (sigma^2)
        kW = kwargs.get("W", None)
        if kW is not None:
            self.W = torch.as_tensor(kW, dtype=config.DTYPE, device=config.DEVICE).flatten()[
                self.mask
            ]
        elif model.target.has_weight:
            self.W = self.model.target[self.fit_window].flatten("weight")[self.mask]
        else:
            self.W = torch.ones_like(self.Y)

        # The forward model which computes the output image given input parameters
        self.forward = lambda x: model(window=self.fit_window, params=x).flatten("data")[self.mask]
        # Compute the jacobian
        self.jacobian = lambda x: model.jacobian(window=self.fit_window, params=x).flatten("data")[
            self.mask
        ]

        # variable to store covariance matrix if it is ever computed
        self._covariance_matrix = None

        # Degrees of freedom
        if ndf is None:
            self.ndf = max(1.0, len(self.Y) - len(self.current_state))
        else:
            self.ndf = ndf

    def chi2_ndf(self):
        return torch.sum(self.W * (self.Y - self.forward(self.current_state)) ** 2) / self.ndf

    @torch.no_grad()
    def fit(self, update_uncertainty=True) -> BaseOptimizer:
        """This performs the fitting operation. It iterates the LM step
        function until convergence is reached. Includes a message
        after fitting to indicate how the fitting exited. Typically if
        the message returns a "success" then the algorithm found a
        minimum. This may be the desired solution, or a pathological
        local minimum, this often depends on the initial conditions.

        """

        if len(self.current_state) == 0:
            if self.verbose > 0:
                config.logger.warning("No parameters to optimize. Exiting fit")
            self.message = "No parameters to optimize. Exiting fit"
            return self

        self._covariance_matrix = None
        self.loss_history = [self.chi2_ndf().item()]
        self.L_history = [self.L]
        self.lambda_history = [self.current_state.detach().clone().cpu().numpy()]
        if self.verbose > 0:
            config.logger.info(
                f"==Starting LM fit for '{self.model.name}' with {len(self.current_state)} dynamic parameters and {len(self.Y)} pixels=="
            )

        for _ in range(self.max_iter):
            if self.verbose > 0:
                config.logger.info(f"Chi^2/DoF: {self.loss_history[-1]:.6g}, L: {self.L:.3g}")
            try:
                if self.fit_valid:
                    with ValidContext(self.model):
                        res = func.lm_step(
                            x=self.model.to_valid(self.current_state),
                            data=self.Y,
                            model=self.forward,
                            weight=self.W,
                            jacobian=self.jacobian,
                            ndf=self.ndf,
                            L=self.L,
                            Lup=self.Lup,
                            Ldn=self.Ldn,
                        )
                    self.current_state = self.model.from_valid(res["x"]).detach()
                else:
                    res = func.lm_step(
                        x=self.current_state,
                        data=self.Y,
                        model=self.forward,
                        weight=self.W,
                        jacobian=self.jacobian,
                        ndf=self.ndf,
                        L=self.L,
                        Lup=self.Lup,
                        Ldn=self.Ldn,
                    )
                    self.current_state = res["x"].detach()
            except OptimizeStopFail:
                if self.verbose > 0:
                    config.logger.warning("Could not find step to improve Chi^2, stopping")
                self.message = self.message + "fail. Could not find step to improve Chi^2"
                break
            except OptimizeStopSuccess as e:
                if self.verbose > 0:
                    config.logger.info(f"Optimization converged successfully: {e}")
                self.message = self.message + "success"
                break

            self.L = np.clip(res["L"], 1e-9, 1e9)
            self.L_history.append(res["L"])
            self.loss_history.append(res["chi2"])
            self.lambda_history.append(self.current_state.detach().clone().cpu().numpy())

            if len(self.loss_history) >= 3:
                if (self.loss_history[-3] - self.loss_history[-1]) / self.loss_history[
                    -1
                ] < self.relative_tolerance and self.L < 0.1:
                    self.message = self.message + "success"
                    break
            if len(self.loss_history) > 10:
                if (self.loss_history[-10] - self.loss_history[-1]) / self.loss_history[
                    -1
                ] < self.relative_tolerance:
                    self.message = (
                        self.message + "success by immobility. Convergence not guaranteed"
                    )
                    break

        else:
            self.message = self.message + "fail. Maximum iterations"

        if self.verbose > 0:
            config.logger.info(
                f"Final Chi^2/DoF: {self.loss_history[-1]:.6g}, L: {self.L_history[-1]:.3g}. Converged: {self.message}"
            )

        self.model.fill_dynamic_values(self.current_state)
        if update_uncertainty:
            self.update_uncertainty()

        return self

    @property
    @torch.no_grad()
    def covariance_matrix(self) -> torch.Tensor:
        """The covariance matrix for the model at the current
        parameters. This can be used to construct a full Gaussian PDF
        for the parameters using: :math:`\\mathcal{N}(\\mu,\\Sigma)`
        where :math:`\\mu` is the optimized parameters and
        :math:`\\Sigma` is the covariance matrix.

        """

        if self._covariance_matrix is not None:
            return self._covariance_matrix
        J = self.jacobian(self.current_state)
        hess = func.hessian(J, self.W)
        try:
            self._covariance_matrix = torch.linalg.inv(hess)
        except:
            config.logger.warning(
                "WARNING: Hessian is singular, likely at least one parameter is non-physical. Will use pseudo-inverse of Hessian to continue but results should be inspected."
            )
            self._covariance_matrix = torch.linalg.pinv(hess)
        return self._covariance_matrix

    @torch.no_grad()
    def update_uncertainty(self) -> None:
        """Call this function after optimization to set the uncertainties for
        the parameters. This will use the diagonal of the covariance
        matrix to update the uncertainties. See the covariance_matrix
        function for the full representation of the uncertainties.

        """
        # set the uncertainty for each parameter
        cov = self.covariance_matrix
        if torch.all(torch.isfinite(cov)):
            try:
                self.model.fill_dynamic_value_uncertainties(torch.sqrt(torch.abs(torch.diag(cov))))
            except RuntimeError as e:
                config.logger.warning(f"Unable to update uncertainty due to: {e}")
        else:
            config.logger.warning(
                "Unable to update uncertainty due to non finite covariance matrix"
            )



# =============================================================================
# Fit all objects identified in a segmentation map
#
# This is a quick script to fit all the objects identified in a segmentation map
# using a single model type. You should set the parameters under PARAMETERS to
# be appropriate for your data. The script will load the target image, mask,
# psf, and variance image (if available) and fit the models to the target image.
#
# First a fit will be run on tight windows exactly enclosing the segmentations
# for each object. Then the windows will be expanded by the set factors and the
# fit will be run again. This is more stable than fitting the expanded windows
# from the start since it reduces the effects of overlap
#
# Run this script with:
# >>> python segmap_models_fit.py
# =============================================================================

import astrophot as ap
import numpy as np
from astropy.io import fits
import matplotlib.pyplot as plt

# PARAMETERS
######################################################################
name = "field_name"  # used for saving files
target_file = "<required>.fits"  # can be a numpy array instead
segmap_file = "<required>.fits"  # can be a numpy array instead
psf_file = None  # "<path to psf>.fits" # can be a numpy array instead
zeropoint = 22.5  # mag
initial_sky = None  # If None, sky will be estimated. Recommended to set manually
sky_locked = False
model_type = "sersic galaxy model"  # model type for segmap entries
segmap_filter = {}  # in pixels or ADU: min_size, min_area, min_flux
segmap_filter_ids = []  # list of segmap ids to remove from fit
segmap_override_init_params = {}  # Override some initial parameters for segmap models
primary_key = None  # segmentation map id, use None to have no primary object
primary_name = "primary object"  # name for primary object
primary_model_type = "spline galaxy model"
primary_initial_params = {}  # {"center": [3, 3], "q": {"value": 0.8, "locked": True}}
# Extra parameters
######################################################################
save_model_image = True
save_residual_image = True
target_hdu = 0  # FITS file index for image data
segmap_hdu = 0
psf_hdu = 0
window_expand_scale = 2  # Windows from segmap will be expanded by this factor
window_expand_border = 10  # Windows from segmap will be expanded by this number of pixels
sky_model_type = "flat sky model"
print_all_models = True
######################################################################

# load target and segmentation map
# ---------------------------------------------------------------------
print("loading target and segmentation map")
target = ap.TargetImage(
    filename=target_file,
    hduext=target_hdu,
    zeropoint=zeropoint,
)

if isinstance(segmap_file, str):
    hdu = fits.open(segmap_file)
    segmap_data = np.array(hdu[segmap_hdu].data, dtype=np.int32)
else:
    segmap_data = segmap_file

# load psf
# ---------------------------------------------------------------------
# PSF
if isinstance(psf_file, str):
    print("loading psf")
    hdu = fits.open(psf_file)
    psf_data = np.array(hdu[psf_hdu].data, dtype=np.float64)
    target.psf = target.psf_image(data=psf_data)
elif psf_file is None:
    psf = None
else:
    target.psf = target.psf_image(data=psf_file)

# Initialization from segmap
# ---------------------------------------------------------------------
print("Parsing segmentaiton map")
windows = ap.utils.initialize.windows_from_segmentation_map(segmap_data)
if len(segmap_filter) > 0:
    windows = ap.utils.initialize.filter_windows(
        windows,
        **segmap_filter,
        image=target,
    )

for ids in segmap_filter_ids:
    del windows[ids]
centers = ap.utils.initialize.centroids_from_segmentation_map(segmap_data, target)
if "galaxy" in model_type:
    PAs = ap.utils.initialize.PA_from_segmentation_map(segmap_data, target, centers)
    qs = ap.utils.initialize.q_from_segmentation_map(segmap_data, target, centers)
else:
    PAs = None
    qs = None
init_params = {}
for window in windows:
    init_params[window] = {"center": centers[window]}
    if "galaxy" in model_type:
        init_params[window]["PA"] = PAs[window]
        init_params[window]["q"] = qs[window]
    init_params[window].update(segmap_override_init_params)

# Create Models
# ---------------------------------------------------------------------
print("Creating models")
models = []
models.append(
    ap.Model(
        name="sky",
        model_type=sky_model_type,
        target=target,
        I=initial_sky if initial_sky is not None else {},
    )
)
if sky_locked:
    models[0].to_static()
primary_model = None
for window in windows:
    if primary_key is not None and window == primary_key:
        print(primary_name, window)
        if "center" not in primary_initial_params:
            primary_initial_params["center"] = init_params[window]["center"]
        if (
            "PA" not in primary_initial_params
            and PAs is not None
            and "galaxy" in primary_model_type
        ):
            primary_initial_params["PA"] = PAs[window]
        if "q" not in primary_initial_params and qs is not None and "galaxy" in primary_model_type:
            primary_initial_params["q"] = qs[window]
        model = ap.Model(
            name=primary_name,
            model_type=primary_model_type,
            target=target,
            **primary_initial_params,
            window=windows[window],
        )
        primary_model = model
    else:
        print(window)
        model = ap.Model(
            name=f"{model_type} {window}",
            model_type=model_type,
            target=target,
            window=windows[window],
            **init_params[window],
        )
    models.append(model)
model = ap.Model(
    name=f"{name} model",
    model_type="group model",
    target=target,
    models=models,
)

# Fit the model
# ---------------------------------------------------------------------
print("Initializing model")
model.initialize()
print("Fitting model round 1")
result = ap.fit.Iter(model, verbose=1).fit()
print("expanding windows")
windows = ap.utils.initialize.scale_windows(
    windows,
    image=target,
    expand_scale=window_expand_scale,
    expand_border=window_expand_border,
)
for i, window in enumerate(windows):
    models[i + 1].window = windows[window]
print("Fitting round 2")
result = ap.fit.Iter(model, verbose=1).fit()

# Report Results
# ----------------------------------------------------------------------
if not sky_locked:
    print(models[0])

if not primary_model is None:
    print(primary_model)
    totmag = primary_model.total_magnitude().detach().cpu().numpy()
    print(f"Total Magnitude: {totmag}")
    if hasattr(primary_model, "radial_model"):
        fig, ax = plt.subplots(figsize=(8, 8))
        ap.plots.radial_light_profile(fig, ax, primary_model)
        plt.savefig(f"{name}_radial_light_profile.jpg")
        plt.close()
    with open(f"{name}_primary_params.csv", "w") as f:
        f.write("Name,Total Magnitude," + ",".join(primary_model.build_params_array_names()) + "\n")
        f.write("string,mag," + ",".join(primary_model.build_params_array_units()) + "\n")
        params = primary_model.build_params_array().detach().cpu().numpy()
        f.write(",".join([str(x) for x in params]) + "\n")

if print_all_models:
    print(model)
    segmap_params = []
    for segmodel in models[1:]:
        if segmodel.name == primary_name:
            continue
        totmag = segmodel.total_magnitude().detach().cpu().numpy()
        segmap_params.append(
            [segmodel.name, totmag] + list(segmodel.build_params_array().detach().cpu().numpy())
        )
    with open(f"{name}_segmap_params.csv", "w") as f:
        f.write("Name,Total Magnitude," + ",".join(segmodel.build_params_array_names()) + "\n")
        f.write("string,mag," + ",".join(segmodel.build_params_array_units()) + "\n")
        for row in segmap_params:
            f.write(",".join([str(x) for x in row]) + "\n")

model.save_state(f"{name}_parameters.hdf5")
if save_model_image:
    model().save(f"{name}_model_image.fits")
    fig, ax = plt.subplots()
    ap.plots.model_image(fig, ax, model)
    plt.savefig(f"{name}_model_image.jpg")
    plt.close()
if save_residual_image:
    (target - model()).save(f"{name}_residual_image.fits")
    fig, ax = plt.subplots()
    ap.plots.residual_image(fig, ax, model, normalize_residuals=True)
    plt.savefig(f"{name}_residual_image.jpg")
    plt.close()



from .lm import LM
from .gradient import Grad
from .iterative import Iter
from .scipy_fit import ScipyFit
from .minifit import MiniFit
from .hmc import HMC
from .mhmcmc import MHMCMC

__all__ = ["LM", "Grad", "Iter", "ScipyFit", "MiniFit", "HMC", "MHMCMC"]



import torch

from ...param import forward
from ...utils.decorators import ignore_numpy_warnings
from .._shared_methods import parametric_initialize, parametric_segment_initialize
from ...utils.parametric_profiles import exponential_np
from .. import func


def _x0_func(model_params, R, F):
    return R[4], 10 ** F[4]


class ExponentialMixin:
    """Exponential radial light profile.

    An exponential is a classical radial model used in many contexts. The
    functional form of the exponential profile is defined as:

    $$I(R) = I_e * \\exp(- b_1(\\frac{R}{R_e} - 1))$$

    Ie is the brightness at the effective radius, and Re is the effective
    radius. `b_1` is a constant that ensures `Ie` is the brightness at `R_e`.

    Parameters:
        Re: effective radius in arcseconds
        Ie: effective surface density in flux/arcsec^2
    """

    _model_type = "exponential"
    _parameter_specs = {
        "Re": {"units": "arcsec", "valid": (0, None), "shape": ()},
        "Ie": {"units": "flux/arcsec^2", "valid": (0, None), "shape": ()},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_initialize(
            self,
            self.target[self.window],
            exponential_np,
            ("Re", "Ie"),
            _x0_func,
        )

    @forward
    def radial_model(self, R, Re, Ie):
        return func.exponential(R, Re, Ie)


class iExponentialMixin:
    """Exponential radial light profile.

    An exponential is a classical radial model used in many contexts. The
    functional form of the exponential profile is defined as:

    $$I(R) = I_e * \\exp(- b_1(\\frac{R}{R_e} - 1))$$

    Ie is the brightness at the effective radius, and Re is the effective
    radius. `b_1` is a constant that ensures `Ie` is the brightness at `R_e`.

    `Re` and `Ie` are batched by their first dimension, allowing for multiple
    exponential profiles to be defined at once.

    Parameters:
        Re: effective radius in arcseconds
        Ie: effective surface density in flux/arcsec^2
    """

    _model_type = "exponential"
    _parameter_specs = {
        "Re": {"units": "arcsec", "valid": (0, None)},
        "Ie": {"units": "flux/arcsec^2", "valid": (0, None)},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_segment_initialize(
            model=self,
            target=self.target[self.window],
            prof_func=exponential_np,
            params=("Re", "Ie"),
            x0_func=_x0_func,
            segments=self.segments,
        )

    @forward
    def iradial_model(self, i, R, Re, Ie):
        return func.exponential(R, Re[i], Ie[i])



import numpy as np


def center_of_mass(image):
    """Determines the light weighted center of mass"""
    ii, jj = np.meshgrid(np.arange(image.shape[0]), np.arange(image.shape[1]), indexing="ij")
    center = np.array((np.sum(image * ii), np.sum(image * jj))) / np.sum(image)
    return center


def recursive_center_of_mass(image, max_iter=10, tol=1e-1):
    """Determines the light weighted center of mass in a progressively smaller window each time centered on the previous center."""

    center = center_of_mass(image)
    for i in range(max_iter):
        width = (image.shape[0] / (3 + i), image.shape[1] / (3 + i))
        ranges = (
            slice(
                max(0, int(center[0] - width[0])), min(image.shape[0], int(center[0] + width[0]))
            ),
            slice(
                max(0, int(center[1] - width[1])), min(image.shape[1], int(center[1] + width[1]))
            ),
        )
        subimage = image[ranges]
        if subimage.size < 9:
            return center
        new_center = center_of_mass(subimage)
        new_center += np.array((ranges[0].start, ranges[1].start))

        if np.linalg.norm(new_center - center) < tol:
            return new_center

        center = new_center
    return center



from typing import List, Optional

import torch
import numpy as np

from .image_object import Image
from .jacobian_image import JacobianImage
from .. import config
from .mixins import DataMixin

__all__ = ["PSFImage"]


class PSFImage(DataMixin, Image):
    """Image object which represents a model of PSF (Point Spread Function).

    PSF_Image inherits from the base Image class and represents the model of a point spread function.
    The point spread function characterizes the response of an imaging system to a point source or point object.

    The shape of the PSF data must be odd.

    Attributes:
        data (torch.Tensor): The image data of the PSF.
        identity (str): The identity of the image. Default is None.

    Methods:
        psf_border_int: Calculates and returns the convolution border size of the PSF image in integer format.
        psf_border: Calculates and returns the convolution border size of the PSF image in the units of pixelscale.
        _save_image_list: Saves the image list to the PSF HDU header.
        reduce: Reduces the size of the image using a given scale factor.
    """

    def __init__(self, *args, **kwargs):
        kwargs.update({"crval": (0, 0), "crpix": (0, 0), "crtan": (0, 0)})
        super().__init__(*args, **kwargs)
        self.crpix = (np.array(self.data.shape, dtype=np.float64) - 1.0) / 2

    def normalize(self):
        """Normalizes the PSF image to have a sum of 1."""
        norm = torch.sum(self.data)
        self.data = self.data / norm
        if self.has_weight:
            self.weight = self.weight * norm**2

    @property
    def psf_pad(self):
        return max(self.data.shape) // 2

    def jacobian_image(
        self,
        parameters: Optional[List[str]] = None,
        data: Optional[torch.Tensor] = None,
        **kwargs,
    ):
        """
        Construct a blank `Jacobian_Image` object formatted like this current `PSF_Image` object. Mostly used internally.
        """
        if parameters is None:
            data = None
            parameters = []
        elif data is None:
            data = torch.zeros(
                (*self.data.shape, len(parameters)),
                dtype=config.DTYPE,
                device=config.DEVICE,
            )
        kwargs = {
            "CD": self.CD.value,
            "crpix": self.crpix,
            "crtan": self.crtan.value,
            "crval": self.crval.value,
            "zeropoint": self.zeropoint,
            "identity": self.identity,
            **kwargs,
        }
        return JacobianImage(parameters=parameters, data=data, **kwargs)

    def model_image(self, **kwargs):
        """
        Construct a blank `Model_Image` object formatted like this current `Target_Image` object. Mostly used internally.
        """
        kwargs = {
            "data": torch.zeros_like(self.data),
            "CD": self.CD.value,
            "crpix": self.crpix,
            "crtan": self.crtan.value,
            "crval": self.crval.value,
            "identity": self.identity,
            **kwargs,
        }
        return PSFImage(**kwargs)

    @property
    def zeropoint(self):
        return None

    @zeropoint.setter
    def zeropoint(self, value):
        """PSFImage does not support zeropoint."""
        pass

    def plane_to_world(self, x, y):
        raise NotImplementedError(
            "PSFImage does not support plane_to_world conversion. There is no meaningful world position of a PSF image."
        )

    def world_to_plane(self, ra, dec):
        raise NotImplementedError(
            "PSFImage does not support world_to_plane conversion. There is no meaningful world position of a PSF image."
        )



import torch

from ...param import forward
from ...utils.decorators import ignore_numpy_warnings
from .._shared_methods import parametric_initialize, parametric_segment_initialize
from ...utils.parametric_profiles import nuker_np
from .. import func


def _x0_func(model_params, R, F):
    return R[4], 10 ** F[4], 1.0, 2.0, 0.5


class NukerMixin:
    """Nuker radial light profile (Lauer et al. 1995).

    This is a classic profile used widely in galaxy modelling. The functional
    form of the Nuker profile is defined as:

    $$I(R) = I_b2^{\\frac{\\beta - \\gamma}{\\alpha}}\\left(\\frac{R}{R_b}\\right)^{-\\gamma}\\left[1 + \\left(\\frac{R}{R_b}\\right)^{\\alpha}\\right]^{\\frac{\\gamma-\\beta}{\\alpha}}$$

    It is effectively a double power law profile. $\\gamma$ gives the inner
    slope, $\\beta$ gives the outer slope, $\\alpha$ is somewhat degenerate with
    the other slopes.

    Parameters:
        Rb: scale length radius
        Ib: intensity at the scale length
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope
    """

    _model_type = "nuker"
    _parameter_specs = {
        "Rb": {"units": "arcsec", "valid": (0, None), "shape": ()},
        "Ib": {"units": "flux/arcsec^2", "valid": (0, None), "shape": ()},
        "alpha": {"units": "none", "valid": (0, None), "shape": ()},
        "beta": {"units": "none", "valid": (0, None), "shape": ()},
        "gamma": {"units": "none", "shape": ()},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_initialize(
            self,
            self.target[self.window],
            nuker_np,
            ("Rb", "Ib", "alpha", "beta", "gamma"),
            _x0_func,
        )

    @forward
    def radial_model(self, R, Rb, Ib, alpha, beta, gamma):
        return func.nuker(R, Rb, Ib, alpha, beta, gamma)


class iNukerMixin:
    """Nuker radial light profile (Lauer et al. 1995).

    This is a classic profile used widely in galaxy modelling. The functional
    form of the Nuker profile is defined as:

    $$I(R) = I_b2^{\\frac{\\beta - \\gamma}{\\alpha}}\\left(\\frac{R}{R_b}\\right)^{-\\gamma}\\left[1 + \\left(\\frac{R}{R_b}\\right)^{\\alpha}\\right]^{\\frac{\\gamma-\\beta}{\\alpha}}$$

    It is effectively a double power law profile. $\\gamma$ gives the inner
    slope, $\\beta$ gives the outer slope, $\\alpha$ is somewhat degenerate with
    the other slopes.

    `Rb`, `Ib`, `alpha`, `beta`, and `gamma` are batched by their first
    dimension, allowing for multiple Nuker profiles to be defined at once.

    Parameters:
        Rb: scale length radius
        Ib: intensity at the scale length
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope
    """

    _model_type = "nuker"
    _parameter_specs = {
        "Rb": {"units": "arcsec", "valid": (0, None)},
        "Ib": {"units": "flux/arcsec^2", "valid": (0, None)},
        "alpha": {"units": "none", "valid": (0, None)},
        "beta": {"units": "none", "valid": (0, None)},
        "gamma": {"units": "none"},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_segment_initialize(
            model=self,
            target=self.target[self.window],
            prof_func=nuker_np,
            params=("Rb", "Ib", "alpha", "beta", "gamma"),
            x0_func=_x0_func,
            segments=self.segments,
        )

    @forward
    def iradial_model(self, i, R, Rb, Ib, alpha, beta, gamma):
        return func.nuker(R, Rb[i], Ib[i], alpha[i], beta[i], gamma[i])



from typing import Union

import torch

from ..image_object import Image
from ..window import Window
from .. import func
from ...utils.interpolate import interp2d
from ...param import forward


class SIPMixin:

    expect_ctype = (("RA---TAN-SIP",), ("DEC--TAN-SIP",))

    def __init__(
        self,
        *args,
        sipA={},
        sipB={},
        sipAP={},
        sipBP={},
        pixel_area_map=None,
        distortion_ij=None,
        distortion_IJ=None,
        filename=None,
        **kwargs,
    ):
        super().__init__(*args, filename=filename, **kwargs)
        if filename is not None:
            return
        self.sipA = sipA
        self.sipB = sipB
        self.sipAP = sipAP
        self.sipBP = sipBP

        self.update_distortion_model(
            distortion_ij=distortion_ij, distortion_IJ=distortion_IJ, pixel_area_map=pixel_area_map
        )

    @forward
    def pixel_to_plane(self, i, j, crtan, CD):
        di = interp2d(self.distortion_ij[0], j, i, padding_mode="border")
        dj = interp2d(self.distortion_ij[1], j, i, padding_mode="border")
        return func.pixel_to_plane_linear(i + di, j + dj, *self.crpix, CD, *crtan)

    @forward
    def plane_to_pixel(self, x, y, crtan, CD):
        I, J = func.plane_to_pixel_linear(x, y, *self.crpix, CD, *crtan)
        dI = interp2d(self.distortion_IJ[0], J, I, padding_mode="border")
        dJ = interp2d(self.distortion_IJ[1], J, I, padding_mode="border")
        return I + dI, J + dJ

    @property
    def pixel_area_map(self):
        return self._pixel_area_map

    def update_distortion_model(self, distortion_ij=None, distortion_IJ=None, pixel_area_map=None):
        """
        Update the pixel area map based on the current SIP coefficients.
        """

        # Pixelized distortion model
        #############################################################
        if distortion_ij is None or distortion_IJ is None:
            i, j = self.pixel_center_meshgrid()
            u, v = i - self.crpix[0], j - self.crpix[1]
            if distortion_ij is None:
                distortion_ij = torch.stack(func.sip_delta(u, v, self.sipA, self.sipB), dim=0)
            if distortion_IJ is None:
                # fixme maybe
                distortion_IJ = torch.stack(func.sip_delta(u, v, self.sipAP, self.sipBP), dim=0)
        self.distortion_ij = distortion_ij
        self.distortion_IJ = distortion_IJ

        # Pixel area map
        #############################################################
        if pixel_area_map is not None:
            self._pixel_area_map = pixel_area_map
            return
        i, j = self.pixel_corner_meshgrid()
        x, y = self.pixel_to_plane(i, j)

        # Shoelace formula for pixel area
        # 1: [:-1, :-1]
        # 2: [:-1, 1:]
        # 3: [1:, 1:]
        # 4: [1:, :-1]
        A = 0.5 * (
            x[:-1, :-1] * y[:-1, 1:]
            + x[:-1, 1:] * y[1:, 1:]
            + x[1:, 1:] * y[1:, :-1]
            + x[1:, :-1] * y[:-1, :-1]
            - (
                x[:-1, 1:] * y[:-1, :-1]
                + x[1:, 1:] * y[:-1, 1:]
                + x[1:, :-1] * y[1:, 1:]
                + x[:-1, :-1] * y[1:, :-1]
            )
        )
        self._pixel_area_map = A.abs()

    def copy(self, **kwargs):
        kwargs = {
            "sipA": self.sipA,
            "sipB": self.sipB,
            "sipAP": self.sipAP,
            "sipBP": self.sipBP,
            "pixel_area_map": self.pixel_area_map,
            **kwargs,
        }
        return super().copy(**kwargs)

    def blank_copy(self, **kwargs):
        kwargs = {
            "sipA": self.sipA,
            "sipB": self.sipB,
            "sipAP": self.sipAP,
            "sipBP": self.sipBP,
            "pixel_area_map": self.pixel_area_map,
            **kwargs,
        }
        return super().blank_copy(**kwargs)

    def get_window(self, other: Union[Image, Window], indices=None, **kwargs):
        """Get a sub-region of the image as defined by an other image on the sky."""
        if indices is None:
            indices = self.get_indices(other if isinstance(other, Window) else other.window)
        return super().get_window(
            other,
            pixel_area_map=self.pixel_area_map[indices],
            indices=indices,
            **kwargs,
        )

    def fits_info(self):
        info = super().fits_info()
        info["CTYPE1"] = "RA---TAN-SIP"
        info["CTYPE2"] = "DEC--TAN-SIP"
        a_order = 0
        for a, b in self.sipA:
            info[f"A_{a}_{b}"] = self.sipA[(a, b)]
            a_order = max(a_order, a + b)
        info["A_ORDER"] = a_order
        b_order = 0
        for a, b in self.sipB:
            info[f"B_{a}_{b}"] = self.sipB[(a, b)]
            b_order = max(b_order, a + b)
        info["B_ORDER"] = b_order
        ap_order = 0
        for a, b in self.sipAP:
            info[f"AP_{a}_{b}"] = self.sipAP[(a, b)]
            ap_order = max(ap_order, a + b)
        info["AP_ORDER"] = ap_order
        bp_order = 0
        for a, b in self.sipBP:
            info[f"BP_{a}_{b}"] = self.sipBP[(a, b)]
            bp_order = max(bp_order, a + b)
        info["BP_ORDER"] = bp_order
        return info

    def load(self, filename: str, hduext=0):
        hdulist = super().load(filename, hduext=hduext)
        self.sipA = {}
        if "A_ORDER" in hdulist[hduext].header:
            a_order = hdulist[hduext].header["A_ORDER"]
            for i in range(a_order + 1):
                for j in range(a_order + 1 - i):
                    key = (i, j)
                    if f"A_{i}_{j}" in hdulist[hduext].header:
                        self.sipA[key] = hdulist[hduext].header[f"A_{i}_{j}"]
        self.sipB = {}
        if "B_ORDER" in hdulist[hduext].header:
            b_order = hdulist[hduext].header["B_ORDER"]
            for i in range(b_order + 1):
                for j in range(b_order + 1 - i):
                    key = (i, j)
                    if f"B_{i}_{j}" in hdulist[hduext].header:
                        self.sipB[key] = hdulist[hduext].header[f"B_{i}_{j}"]
        self.sipAP = {}
        if "AP_ORDER" in hdulist[hduext].header:
            ap_order = hdulist[hduext].header["AP_ORDER"]
            for i in range(ap_order + 1):
                for j in range(ap_order + 1 - i):
                    key = (i, j)
                    if f"AP_{i}_{j}" in hdulist[hduext].header:
                        self.sipAP[key] = hdulist[hduext].header[f"AP_{i}_{j}"]
        self.sipBP = {}
        if "BP_ORDER" in hdulist[hduext].header:
            bp_order = hdulist[hduext].header["BP_ORDER"]
            for i in range(bp_order + 1):
                for j in range(bp_order + 1 - i):
                    key = (i, j)
                    if f"BP_{i}_{j}" in hdulist[hduext].header:
                        self.sipBP[key] = hdulist[hduext].header[f"BP_{i}_{j}"]
        self.update_distortion_model()
        return hdulist
